"""Typing."""
from __future__ import annotations
from typing import Union, Any, Mapping, Sequence, List, Tuple, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color

ColorInput = Union['Color', str, Mapping[str, Any]]

# Vectors, Matrices, and Arrays are assumed to be mutable lists
Vector = List[float]
Matrix = List[Vector]
Array = Union[Matrix, Vector]

# Anything that resembles a sequence will be considered "like" one of our types above
VectorLike = Sequence[float]
MatrixLike = Sequence[VectorLike]
ArrayLike = Union[VectorLike, MatrixLike]

# General algebra types
Shape = Tuple[int, ...]
ShapeLike = Sequence[int]
DimHints = Tuple[int, int]

# For times when we must explicitly say we support `int` and `float`
SupportsFloatOrInt = TypeVar('SupportsFloatOrInt', float, int)


class Plugin:
    """
    Plugin type base class.

    A common class used to help simplify typing in some cases.
    """

    NAME = ""



"""Channels."""
from __future__ import annotations

FLG_ANGLE = 1
FLG_PERCENT = 2
FLG_OPT_PERCENT = 4
FLG_MIRROR_PERCENT = 8


class Channel(str):
    """Channel."""

    low: float
    high: float
    span: float
    offset: float
    bound: bool
    flags: int
    limit: tuple[float | None, float | None]
    nans: float

    def __new__(
        cls,
        name: str,
        low: float,
        high: float,
        mirror_range: bool = False,
        bound: bool = False,
        flags: int = 0,
        limit: tuple[float | None, float | None] = (None, None),
        nans: float = 0.0
    ) -> 'Channel':
        """Initialize."""

        obj = super().__new__(cls, name)
        obj.low = low
        obj.high = high
        mirror = flags & FLG_MIRROR_PERCENT and abs(low) == high
        obj.span = high if mirror else high - low
        obj.offset = 0.0 if mirror else -low
        obj.bound = bound
        obj.flags = flags
        obj.limit = limit
        obj.nans = nans

        return obj



"""Temperature plugin."""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from ..types import Plugin, Vector
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class CCT(Plugin, metaclass=ABCMeta):
    """Delta E plugin class."""

    NAME = ''

    @abstractmethod
    def to_cct(self, color: Color, **kwargs: Any) -> Vector:
        """Calculate a color's CCT."""

    @abstractmethod
    def from_cct(
        self,
        color: type[Color],
        space: str,
        kelvin: float,
        duv: float,
        scale: bool,
        scale_space: str | None,
        **kwargs: Any
    ) -> Color:
        """Calculate a color that satisfies the CCT."""


def cct(name: str | None, color: type[Color] | Color) -> CCT:
    """Get the appropriate contrast plugin."""

    if name is None:
        name = color.CCT

    method = color.CCT_MAP.get(name)
    if not method:
        raise ValueError("'{}' CCT method is not supported".format(name))

    return method



"""Test CAM16 UCS."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestCAM16CAM16UCS(util.ColorAssertsPyTest):
    """Test CAM16 UCS."""

    COLORS = [
        ('red', 'color(--cam16-ucs 59.178 40.82 21.153)'),
        ('orange', 'color(--cam16-ucs 78.364 9.6945 28.629)'),
        ('yellow', 'color(--cam16-ucs 96.802 -12.779 33.037)'),
        ('green', 'color(--cam16-ucs 46.661 -26.746 20.671)'),
        ('blue', 'color(--cam16-ucs 36.252 8.5723 -37.87)'),
        ('indigo', 'color(--cam16-ucs 24.524 19.714 -22.758)'),
        ('violet', 'color(--cam16-ucs 74.738 27.949 -15.247)'),
        ('white', 'color(--cam16-ucs 100 -1.8983 -1.0754)'),
        ('gray', 'color(--cam16-ucs 56.23 -1.2555 -0.71134)'),
        ('black', 'color(--cam16-ucs 0 0 0)'),
        # Test color
        ('color(--cam16-ucs 50 10 -10)', 'color(--cam16-ucs 50 10 -10)'),
        ('color(--cam16-ucs 50 10 -10 / 0.5)', 'color(--cam16-ucs 50 10 -10 / 0.5)'),
        ('color(--cam16-ucs 50% 50% -50% / 50%)', 'color(--cam16-ucs 50 25 -25 / 0.5)'),
        ('color(--cam16-ucs none none none / none)', 'color(--cam16-ucs none none none / none)'),
        # Test percent ranges
        ('color(--cam16-ucs 0% 0% 0%)', 'color(--cam16-ucs 0 0 0)'),
        ('color(--cam16-ucs 100% 100% 100%)', 'color(--cam16-ucs 100 50 50)'),
        ('color(--cam16-ucs -100% -100% -100%)', 'color(--cam16-ucs -100 -50 -50)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cam16-ucs'), Color(color2))


class TestCAM16UCSSerialize(util.ColorAssertsPyTest):
    """Test CAM16 UCS serialization."""

    COLORS = [
        # Test color
        ('color(--cam16-ucs 75 10 -10 / 0.5)', {}, 'color(--cam16-ucs 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--cam16-ucs 75 10 -10)', {'alpha': True}, 'color(--cam16-ucs 75 10 -10 / 1)'),
        ('color(--cam16-ucs 75 10 -10 / 0.5)', {'alpha': False}, 'color(--cam16-ucs 75 10 -10)'),
        # Test None
        ('color(--cam16-ucs none 10 -10)', {}, 'color(--cam16-ucs 0 10 -10)'),
        ('color(--cam16-ucs none 10 -10)', {'none': True}, 'color(--cam16-ucs none 10 -10)'),
        # Test Fit (not bound)
        ('color(--cam16-ucs 120 10 -10)', {}, 'color(--cam16-ucs 120 10 -10)'),
        ('color(--cam16-ucs 120 10 -10)', {'fit': False}, 'color(--cam16-ucs 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestCAM16UCSPoperties(util.ColorAsserts, unittest.TestCase):
    """Test CAM16 UCS."""

    def test_j(self):
        """Test `j`."""

        c = Color('color(--cam16-ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['j'], 0.51332)
        c['j'] = 0.2
        self.assertEqual(c['j'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--cam16-ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['a'], 0.92781)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--cam16-ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['b'], 1.076)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cam16-ucs 0.51332 0.92781 1.076)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestCAM16CAM16SCD(util.ColorAssertsPyTest):
    """Test CAM16 SCD."""

    COLORS = [
        ('red', 'color(--cam16-scd 59.178 33.597 17.41)'),
        ('orange', 'color(--cam16-scd 78.364 8.3723 24.725)'),
        ('yellow', 'color(--cam16-scd 96.802 -10.847 28.041)'),
        ('green', 'color(--cam16-scd 46.661 -22.82 17.637)'),
        ('blue', 'color(--cam16-scd 36.252 7.1995 -31.806)'),
        ('indigo', 'color(--cam16-scd 24.524 17.032 -19.663)'),
        ('violet', 'color(--cam16-scd 74.738 24.003 -13.094)'),
        ('white', 'color(--cam16-scd 100 -1.8713 -1.0602)'),
        ('gray', 'color(--cam16-scd 56.23 -1.2436 -0.70457)'),
        ('black', 'color(--cam16-scd 0 0 0)'),
        # Test color
        ('color(--cam16-scd 50 10 -10)', 'color(--cam16-scd 50 10 -10)'),
        ('color(--cam16-scd 50 10 -10 / 0.5)', 'color(--cam16-scd 50 10 -10 / 0.5)'),
        ('color(--cam16-scd 50% 50% -50% / 50%)', 'color(--cam16-scd 50 20 -20 / 0.5)'),
        ('color(--cam16-scd none none none / none)', 'color(--cam16-scd none none none / none)'),
        # Test percent ranges
        ('color(--cam16-scd 0% 0% 0%)', 'color(--cam16-scd 0 0 0)'),
        ('color(--cam16-scd 100% 100% 100%)', 'color(--cam16-scd 100 40 40)'),
        ('color(--cam16-scd -100% -100% -100%)', 'color(--cam16-scd -100 -40 -40)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cam16-scd'), Color(color2))


class TestCAM16SCDSerialize(util.ColorAssertsPyTest):
    """Test CAM16 SCD serialization."""

    COLORS = [
        # Test color
        ('color(--cam16-scd 75 10 -10 / 0.5)', {}, 'color(--cam16-scd 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--cam16-scd 75 10 -10)', {'alpha': True}, 'color(--cam16-scd 75 10 -10 / 1)'),
        ('color(--cam16-scd 75 10 -10 / 0.5)', {'alpha': False}, 'color(--cam16-scd 75 10 -10)'),
        # Test None
        ('color(--cam16-scd none 10 -10)', {}, 'color(--cam16-scd 0 10 -10)'),
        ('color(--cam16-scd none 10 -10)', {'none': True}, 'color(--cam16-scd none 10 -10)'),
        # Test Fit (not bound)
        ('color(--cam16-scd 120 10 -10)', {}, 'color(--cam16-scd 120 10 -10)'),
        ('color(--cam16-scd 120 10 -10)', {'fit': False}, 'color(--cam16-scd 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestCAM16SCDPoperties(util.ColorAsserts, unittest.TestCase):
    """Test CAM16 SCD."""

    def test_j(self):
        """Test `j`."""

        c = Color('color(--cam16-scd 0.51332 0.92781 1.076)')
        self.assertEqual(c['j'], 0.51332)
        c['j'] = 0.2
        self.assertEqual(c['j'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--cam16-scd 0.51332 0.92781 1.076)')
        self.assertEqual(c['a'], 0.92781)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--cam16-scd 0.51332 0.92781 1.076)')
        self.assertEqual(c['b'], 1.076)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cam16-scd 0.51332 0.92781 1.076)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestCAM16CAM16LCD(util.ColorAssertsPyTest):
    """Test CAM16 LCD."""

    COLORS = [
        ('red', 'color(--cam16-lcd 59.178 59.994 31.089)'),
        ('orange', 'color(--cam16-lcd 78.364 12.557 37.084)'),
        ('yellow', 'color(--cam16-lcd 96.802 -17.273 44.653)'),
        ('green', 'color(--cam16-lcd 46.661 -35.677 27.573)'),
        ('blue', 'color(--cam16-lcd 36.252 11.909 -52.61)'),
        ('indigo', 'color(--cam16-lcd 24.524 25.511 -29.451)'),
        ('violet', 'color(--cam16-lcd 74.738 36.686 -20.013)'),
        ('white', 'color(--cam16-lcd 100 -1.9348 -1.0961)'),
        ('gray', 'color(--cam16-lcd 56.23 -1.2714 -0.72038)'),
        ('black', 'color(--cam16-lcd 0 0 0)'),
        # Test color
        ('color(--cam16-lcd 50 10 -10)', 'color(--cam16-lcd 50 10 -10)'),
        ('color(--cam16-lcd 50 10 -10 / 0.5)', 'color(--cam16-lcd 50 10 -10 / 0.5)'),
        ('color(--cam16-lcd 50% 50% -50% / 50%)', 'color(--cam16-lcd 50 35 -35 / 0.5)'),
        ('color(--cam16-lcd none none none / none)', 'color(--cam16-lcd none none none / none)'),
        # Test percent ranges
        ('color(--cam16-lcd 0% 0% 0%)', 'color(--cam16-lcd 0 0 0)'),
        ('color(--cam16-lcd 100% 100% 100%)', 'color(--cam16-lcd 100 70 70)'),
        ('color(--cam16-lcd -100% -100% -100%)', 'color(--cam16-lcd -100 -70 -70)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cam16-lcd'), Color(color2))


class TestCAM16LCDSerialize(util.ColorAssertsPyTest):
    """Test CAM16 LCD serialization."""

    COLORS = [
        # Test color
        ('color(--cam16-lcd 75 10 -10 / 0.5)', {}, 'color(--cam16-lcd 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--cam16-lcd 75 10 -10)', {'alpha': True}, 'color(--cam16-lcd 75 10 -10 / 1)'),
        ('color(--cam16-lcd 75 10 -10 / 0.5)', {'alpha': False}, 'color(--cam16-lcd 75 10 -10)'),
        # Test None
        ('color(--cam16-lcd none 10 -10)', {}, 'color(--cam16-lcd 0 10 -10)'),
        ('color(--cam16-lcd none 10 -10)', {'none': True}, 'color(--cam16-lcd none 10 -10)'),
        # Test Fit (not bound)
        ('color(--cam16-lcd 120 10 -10)', {}, 'color(--cam16-lcd 120 10 -10)'),
        ('color(--cam16-lcd 120 10 -10)', {'fit': False}, 'color(--cam16-lcd 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestCAM16LCDPoperties(util.ColorAsserts, unittest.TestCase):
    """Test CAM16 LCD."""

    def test_j(self):
        """Test `j`."""

        c = Color('color(--cam16-lcd 0.51332 0.92781 1.076)')
        self.assertEqual(c['j'], 0.51332)
        c['j'] = 0.2
        self.assertEqual(c['j'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--cam16-lcd 0.51332 0.92781 1.076)')
        self.assertEqual(c['a'], 0.92781)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--cam16-lcd 0.51332 0.92781 1.076)')
        self.assertEqual(c['b'], 1.076)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cam16-lcd 0.51332 0.92781 1.076)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('cam16').is_achromatic(), True)
        self.assertEqual(Color('srgb', [0.000000001] * 3).convert('cam16-ucs').set('j', NaN).is_achromatic(), True)
        self.assertEqual(Color('cam16-ucs', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('cam16-ucs', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('cam16-ucs', [0, 3, -4]).is_achromatic(), True)
        self.assertEqual(Color('cam16-ucs', [NaN, 0, -3]).is_achromatic(), True)
        self.assertEqual(Color('cam16-ucs', [30, NaN, 0]).is_achromatic(), False)
        self.assertEqual(Color('cam16-ucs', [NaN, NaN, 0]).is_achromatic(), True)



"""Test gamut mapping fitting."""
import unittest
import math
from coloraide.everything import ColorAll as Color
from coloraide import gamut
from . import util


class TestGamut(util.ColorAsserts, unittest.TestCase):
    """Test gamut mapping/fitting."""

    def test_in_gamut(self):
        """Test in gamut check."""

        self.assertTrue(Color('red').in_gamut())

    def test_out_of_gamut(self):
        """Test in gamut check."""

        self.assertFalse(Color('color(srgb 2 0 0)').in_gamut())

    def test_in_gamut_other_space(self):
        """Test if a color is in gamut in a different space."""

        self.assertTrue(Color('red').convert('lch').in_gamut('srgb'))

    def test_fit(self):
        """Test fit."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        color2 = color.clone().fit()
        self.assertIsNot(color, color2)
        self.assertTrue(color2.in_gamut())

    def test_fit_clip(self):
        """Test fit."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        color2 = color.clone().fit(method="clip")
        self.assertIsNot(color, color2)
        self.assertTrue(color2.in_gamut())
        color3 = color.clone().fit()
        self.assertColorNotEqual(color2, color3)

    def test_clip(self):
        """Test fit."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        color2 = color.clone().clip()
        self.assertIsNot(color, color2)
        self.assertTrue(color2.in_gamut())
        color3 = color.clone().fit()
        self.assertColorNotEqual(color2, color3)

    def test_clip_in_place(self):
        """Test clip in place."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        color2 = color.clip()
        self.assertIs(color, color2)
        self.assertTrue(color2.in_gamut())
        color3 = color.clone().fit()
        self.assertColorEqual(color2, color3)

    def test_clip_other_space(self):
        """Test clip other space."""

        color = Color('hsl(330 110% 50%)')
        self.assertFalse(color.in_gamut('srgb'))
        self.assertTrue(color.clip('srgb').in_gamut('srgb'))

    def test_clip_undefined(self):
        """Test that clip leaves undefined channels alone."""

        color = Color('rgb(none 255.1 -0.1)')
        self.assertFalse(color.in_gamut('srgb'))
        color2 = color.clone().clip('srgb')
        self.assertColorEqual(color2, Color('rgb(0 255 0)'))
        self.assertTrue(color2.in_gamut('srgb'))

    def test_sdr_extremes_low(self):
        """Test SDR extreme low case."""

        color = Color('hsl(none 0% -10%)')
        self.assertColorEqual(color.fit(), Color('hsl(0 0% 0%)'))

    def test_sdr_extremes_high(self):
        """Test SDR extreme high case."""

        color = Color('hsl(none 0% 110%)')
        self.assertColorEqual(color.fit(), Color('hsl(0 0% 100%)'))

    def test_hdr_extreme_high(self):
        """Test HDR extreme high case."""

        color = Color('color(--rec2100-pq 1.01 0.2 0)')
        self.assertColorEqual(color.fit(), Color('color(--rec2100-pq 1 0.63544 1)'))

    def test_hdr_extreme_low(self):
        """Test HDR extreme low case."""

        color = Color('color(--rec2100-pq -0.3 -0.2 0)')
        self.assertColorEqual(color.fit(), Color('color(--rec2100-pq 0.0 0.0 0.0)'))

    def test_bad_fit(self):
        """Test fit."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        with self.assertRaises(ValueError):
            color.fit(method="bad")

    def test_fit_in_place(self):
        """Test fit in place."""

        color = Color('color(srgb 2 0.5 0.5)')
        self.assertFalse(color.in_gamut())
        color2 = color.fit()
        self.assertIs(color, color2)
        self.assertTrue(color2.in_gamut())

    def test_fit_other_space(self):
        """Test fit in other space."""

        color = Color('hsl(330 110% 50%)')
        self.assertFalse(color.in_gamut('srgb'))
        self.assertTrue(color.fit('srgb').in_gamut('srgb'))

    def test_out_of_gamut_other_space(self):
        """Test if a color is in gamut in a different space."""

        self.assertFalse(Color('color(srgb 2 0 0)').convert('lch').in_gamut('srgb'))


class TestHCTGamut(util.ColorAsserts, unittest.TestCase):
    """
    Test HCT gamut mapping by producing tonal maps.

    When we are off, we seem to only be off on a given channel by at most < +/- 1 (between 0 - 255 floating point).
    As this is not a 1:1 port of the Material utilities, different precision of matrices and different gamut mapping
    can cause slight differences in results. We are not aiming to match the Material implementation, but aiming to match
    the intent.
    """

    def test_blue(self):
        """Test blue tonal maps."""

        hct = Color('#0000ff').convert('hct')
        tones = [100, 95, 90, 80, 70, 60, 50, 40, 30, 20, 10, 0]
        expect = ['#ffffff', '#f1efff', '#e0e0ff', '#bec2ff',
                  '#9da3ff', '#7c84ff', '#5a64ff', '#343dff',
                  '#0000ef', '#0001ac', '#00006e', '#000000']

        # A channel should be off no more than by 1 in a scale of 0 - 255
        # Due to the fact that Material uses less precise matrices
        for tone, answer in zip(tones, expect):
            s1 = [c * 255 for c in hct.clone().set('tone', tone).convert('srgb').fit(method='hct-chroma')[:-1]]
            s2 = [c * 255 for c in Color(answer)[:-1]]
            for c1, c2 in zip(s1, s2):
                self.assertTrue(math.isclose(c1, c2, abs_tol=1.0))

        hct.set('chroma', 16)
        expect = ['#ffffff', '#f1efff', '#e1e0f9', '#c5c4dd',
                  '#a9a9c1', '#8f8fa6', '#75758b', '#5c5d72',
                  '#444559', '#2e2f42', '#191a2c', '#000000']

        for tone, answer in zip(tones, expect):
            s1 = [c * 255 for c in hct.clone().set('tone', tone).convert('srgb').fit(method='hct-chroma')[:-1]]
            s2 = [c * 255 for c in Color(answer)[:-1]]
            for c1, c2 in zip(s1, s2):
                self.assertTrue(math.isclose(c1, c2, abs_tol=1.0))


class TestPointerGamut(util.ColorAsserts, unittest.TestCase):
    """Test Pointer's gamut."""

    def test_in_pointer_gamut_too_light(self):
        """Test gamut check when too light."""

        self.assertFalse(Color('white').in_pointer_gamut())

    def test_in_pointer_gamut_too_dark(self):
        """Test gamut check when too dark."""

        self.assertFalse(Color('black').in_pointer_gamut())

    def test_in_pointer_gamut(self):
        """Test gamut check in gamut."""

        self.assertTrue(Color('orange').in_pointer_gamut())

    def test_out_of_pointer_gamut(self):
        """Test gamut check out of gamut."""

        self.assertFalse(Color('red').in_pointer_gamut())

    def test_hue_wrap(self):
        """Test when hue interpolation wraps."""

        self.assertFalse(Color('deeppink').in_pointer_gamut())

    def test_fit_pointer_gamut_too_light(self):
        """Test fitting gamut when too light."""

        self.assertColorEqual(Color('white').fit_pointer_gamut(), Color('rgb(226.33 226.33 226.33)'))

    def test_fit_pointer_gamut_too_dark(self):
        """Test fitting gamut when too dark."""

        self.assertColorEqual(Color('black').fit_pointer_gamut(), Color('rgb(37.667 37.667 37.667)'))

    def test_fit_pointer_gamut(self):
        """Test gamut check out of gamut."""

        self.assertColorEqual(Color('red').fit_pointer_gamut(), Color('rgb(244 46.539 23.139)'))

    def test_pointer_boundary_too_light(self):
        """Test when pointer boundary is request is too light."""

        with self.assertRaises(ValueError):
            gamut.pointer.pointer_gamut_boundary(0)

    def test_pointer_boundary_too_dark(self):
        """Test when pointer boundary is request is too dark."""

        with self.assertRaises(ValueError):
            gamut.pointer.pointer_gamut_boundary(100)

    def test_pointer_boundary_mid(self):
        """Test when pointer boundary request."""

        boundary = gamut.pointer.pointer_gamut_boundary(50)
        # Test a sample of the values
        test = [boundary[0], boundary[5], boundary[8]]
        expected = [[0.47971142940104206, 0.2384179501289976, 18.418651851244416],
                    [0.634800855490809, 0.35129621215430396, 18.418651851244416],
                    [0.49892034937725527, 0.45347791373057683, 18.418651851244416]]

        for coords1, coords2 in zip(test, expected):
            [self.assertCompare(a, b) for a, b in zip(coords1, coords2)]

    def test_pointer_boundary_max(self):
        """Test pointer boundary max request."""

        boundary = gamut.pointer.pointer_gamut_boundary()
        # Test a sample of the values
        test = [boundary[0], boundary[5], boundary[8]]
        expected = [[0.5076988072583095, 0.2255929648510513, 11.250973799663784],
                    [0.634800855490809, 0.35129621215430396, 18.418651851244416],
                    [0.5260711281340981, 0.462168144862833, 48.2781043708229]]

        for coords1, coords2 in zip(test, expected):
            [self.assertCompare(a, b) for a, b in zip(coords1, coords2)]



"""
L* color contrast.

Used for color contrast in Google's HCT.

https://material.io/blog/science-of-color-design
"""
from __future__ import annotations
from ..contrast import ColorContrast
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class LstarContrast(ColorContrast):
    """L* contrast."""

    NAME = "lstar"

    def contrast(self, color1: Color, color2: Color, **kwargs: Any) -> float:
        """Contrast."""

        l1 = color1.get('lch-d65.lightness', nans=False)
        l2 = color2.get('lch-d65.lightness', nans=False)

        if l1 > l2:
            l2, l1 = l1, l2

        return l2 - l1



"""
ACES 2065-1 color space.

https://www.oscars.org/science-technology/aces/aces-documentation
"""
from __future__ import annotations
from ..channels import Channel
from ..spaces.srgb import sRGB
from .. import algebra as alg
from ..types import Vector

AP0_TO_XYZ = [
    [0.9525523959381857, 0.0, 9.367863166046855e-05],
    [0.34396644976507507, 0.7281660966134857, -0.07213254637856079],
    [0.0, 0.0, 1.0088251843515859]
]

XYZ_TO_AP0 = [
    [1.0498110174979742, 0.0, -9.748454057925287e-05],
    [-0.49590302307731976, 1.3733130458157063, 0.09824003605730998],
    [0.0, 0.0, 0.991252018200499]
]

MIN = 0.0
MAX = 1.0


def aces_to_xyz(aces: Vector) -> Vector:
    """Convert ACEScc to XYZ."""

    return alg.dot(AP0_TO_XYZ, aces, dims=alg.D2_D1)


def xyz_to_aces(xyz: Vector) -> Vector:
    """Convert XYZ to ACEScc."""

    return alg.dot(XYZ_TO_AP0, xyz, dims=alg.D2_D1)


class ACES20651(sRGB):
    """The ACES color class."""

    BASE = "xyz-d65"
    NAME = "aces2065-1"
    SERIALIZE = ("--aces2065-1",)  # type: tuple[str, ...]
    WHITE = (0.32168, 0.33767)
    CHANNELS = (
        Channel("r", 0.0, 65504.0, bound=True),
        Channel("g", 0.0, 65504.0, bound=True),
        Channel("b", 0.0, 65504.0, bound=True)
    )
    DYNAMIC_RANGE = 'hdr'

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return aces_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return xyz_to_aces(coords)



"""
Rec. 709 color space class.

This color space uses the same chromaticities and white points as sRGB,
but uses the same gamma correction as Rec. 2020, just at 10 bit precision.

- https://en.wikipedia.org/wiki/Rec._709
- https://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.709-6-201506-I!!PDF-E.pdf
"""
from __future__ import annotations
from .srgb import sRGB
import math
from .. import algebra as alg
from ..types import Vector

ALPHA = 1.099
BETA = 0.018
BETA45 = BETA * 4.5
ALPHAM1 = 0.099


def lin_709(rgb: Vector) -> Vector:
    """
    Convert an array of Rec. 709 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    Transfer function is similar to Rec. 2020, just at a lower precision
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA45:
            result.append(i / 4.5)
        else:
            result.append(math.copysign(alg.nth_root((abs_i + ALPHAM1) / ALPHA, 0.45), i))
    return result


def gam_709(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light Rec. 709 RGB  in the range 0.0-1.0 to gamma corrected form.

    Transfer function is similar to Rec. 2020, just at a lower precision
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA:
            result.append(4.5 * i)
        else:
            result.append(math.copysign(ALPHA * (abs_i ** 0.45) - ALPHAM1, i))
    return result


class Rec709(sRGB):
    """Rec. 709 class."""

    BASE = "srgb-linear"
    NAME = "rec709"
    SERIALIZE = ("--rec709",)

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Rec. 709."""

        return lin_709(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Rec. 709."""

        return gam_709(coords)



"""
RLAB.

https://scholarworks.rit.edu/cgi/viewcontent.cgi?article=1153&context=article
https://www.imaging.org/site/PDFS/Papers/1997/RP-0-67/2368.pdf

There was some conflicting info between the two papers. One cited M one way, and one cited the inverse.
One was clearly a mistake. It seems that the `imaging.org` one is the correct M. We ended up implementing
the "Refinement of the RLAB Color Space" algorithm and just replaced M with the correct one.

We did recalculate R for our white point to get a nice, clean transform.
"""

import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import algebra as alg  # noqa: E402
from coloraide import util  # noqa: E402
from coloraide.cat import WHITES  # noqa: E402


white_d65 = util.xy_to_xyz((0.31270, 0.32900))

R = [
    [1.9569, -1.1882, 0.2313],
    [0.3612, 0.6388, 0.0],
    [0.0, 0.0, 1.0000]
]

M = [
    [0.4002, 0.7076, -0.0808],
    [-0.2263, 1.1653, 0.0457],
    [0.0, 0.0, 0.9182]
]

YN = 318.0  # D65 `318 cd / m^2`
D = 0.0


def calc_a_lms(lms):
    """Calculate the various matrix diagonal values."""

    a = []
    s = sum(lms)
    for c in lms:
        l = (3.0 * c) / s
        p = (1.0 + alg.nth_root(YN, 3) + l) / (1.0 + alg.nth_root(YN, 3) + 1.0 / l)
        a.append((p + D * (1.0 - p)) / c)
    return a


def calc_ram(white):
    """Calculate RAM."""

    xyz_w = util.xy_to_xyz(white)
    lms = alg.dot(M, xyz_w)
    a_lms = calc_a_lms(lms)
    A = alg.diag(a_lms)
    R = alg.multi_dot([alg.inv(alg.diag(xyz_w)), alg.inv(M), alg.inv(A)])
    ram = alg.dot(R, alg.dot(A, M))
    return ram


def pprint(value):
    """Print the matrix."""
    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


if __name__ == "__main__":
    RAM = calc_ram(WHITES['2deg']['D65'])
    print('===== XYZ to XYZref =====')
    pprint(RAM)
    print('===== XYZref to XYZ =====')
    pprint(alg.inv(RAM))



"""Test chromaticity methods."""
from coloraide import Color
from . import util
import pytest
import unittest
from coloraide import util as cutil


class TestChromaticity(util.ColorAssertsPyTest):
    """Test chromaticity conversions."""

    @pytest.mark.parametrize(
        'color,xy',
        [
            ('color(srgb 1 0 0)', (0.64, 0.33)),
            ('color(srgb 0 1 0)', (0.30, 0.60)),
            ('color(srgb 0 0 1)', (0.15, 0.06)),
            ('color(display-p3 1 0 0)', (0.68, 0.32)),
            ('color(display-p3 0 1 0)', (0.265, 0.69)),
            ('color(display-p3 0 0 1)', (0.150, 0.060)),
            ('color(a98-rgb 1 0 0)', (0.64, 0.33)),
            ('color(a98-rgb 0 1 0)', (0.21, 0.71)),
            ('color(a98-rgb 0 0 1)', (0.15, 0.06)),
            ('color(prophoto-rgb 1 0 0)', (0.7347, 0.2653)),
            ('color(prophoto-rgb 0 1 0)', (0.1596, 0.8404)),
            ('color(prophoto-rgb 0 0 1)', (0.0366, 0.0001)),
            ('color(rec2020 1 0 0)', (0.708, 0.292)),
            ('color(rec2020 0 1 0)', (0.17, 0.797)),
            ('color(rec2020 0 0 1)', (0.131, 0.046))
        ]
    )
    def test_xy(self, color, xy):
        """Test `xy` conversions."""

        xy2 = Color(color).xy()
        for v1, v2 in zip(xy, xy2):
            self.assertCompare(v1, v2)

    @pytest.mark.parametrize(
        'color,uv',
        [
            ('color(srgb 1 0 0)', (0.4507, 0.52289)),
            ('color(srgb 0 1 0)', (0.125, 0.5625)),
            ('color(srgb 0 0 1)', (0.17544, 0.15789)),
            ('color(display-p3 1 0 0)', (0.49635, 0.52555)),
            ('color(display-p3 0 1 0)', (0.0986, 0.57767)),
            ('color(display-p3 0 0 1)', (0.17544, 0.15789)),
            ('color(a98-rgb 1 0 0)', (0.4507, 0.52289)),
            ('color(a98-rgb 0 1 0)', (0.07568, 0.57568)),
            ('color(a98-rgb 0 0 1)', (0.17544, 0.15789)),
            ('color(prophoto-rgb 1 0 0)', (0.62339, 0.50649)),
            ('color(prophoto-rgb 0 1 0)', (0.05001, 0.5925)),
            ('color(prophoto-rgb 0 0 1)', (0.05, 0.00031)),
            ('color(rec2020 1 0 0)', (0.5566, 0.51651)),
            ('color(rec2020 0 1 0)', (0.05563, 0.5868)),
            ('color(rec2020 0 0 1)', (0.15927, 0.12584))
        ]
    )
    def test_xy_1976(self, color, uv):
        """Test `uv` 1976 conversions."""

        uv2 = Color(color).uv()
        for v1, v2 in zip(uv, uv2):
            self.assertCompare(v1, v2)

    @pytest.mark.parametrize(
        'color,uv',
        [
            ('color(srgb 1 0 0)', (0.4507, 0.34859)),
            ('color(srgb 0 1 0)', (0.125, 0.375)),
            ('color(srgb 0 0 1)', (0.17544, 0.10526)),
            ('color(display-p3 1 0 0)', (0.49635, 0.35036)),
            ('color(display-p3 0 1 0)', (0.0986, 0.38512)),
            ('color(display-p3 0 0 1)', (0.17544, 0.10526)),
            ('color(a98-rgb 1 0 0)', (0.4507, 0.34859)),
            ('color(a98-rgb 0 1 0)', (0.07568, 0.38378)),
            ('color(a98-rgb 0 0 1)', (0.17544, 0.10526)),
            ('color(prophoto-rgb 1 0 0)', (0.62339, 0.33766)),
            ('color(prophoto-rgb 0 1 0)', (0.05001, 0.395)),
            ('color(prophoto-rgb 0 0 1)', (0.05, 0.0002)),
            ('color(rec2020 1 0 0)', (0.5566, 0.34434)),
            ('color(rec2020 0 1 0)', (0.05563, 0.3912)),
            ('color(rec2020 0 0 1)', (0.15927, 0.08389))
        ]
    )
    def test_xy_1960(self, color, uv):
        """Test `uv` 1960 conversions."""

        uv2 = Color(color).uv('1960')
        for v1, v2 in zip(uv, uv2):
            self.assertCompare(v1, v2)


class TestChromaticitySpecificCases(util.ColorAsserts, unittest.TestCase):
    """Test distance specific cases."""

    def test_xy_luminance(self):
        """Test xy with luminance."""

        xyy = Color('red').split_chromaticity('xy-1931')
        for v1, v2 in zip(xyy, [0.64, 0.33, 0.21264]):
            self.assertCompare(v1, v2)

    def test_uv_1960_luminance(self):
        """Test uv 1960 with luminance."""

        uvy = Color('red').split_chromaticity('uv-1960')
        for v1, v2 in zip(uvy, [0.4507, 0.34859, 0.21264]):
            self.assertCompare(v1, v2)

    def test_uv_1976_luminance(self):
        """Test uv 1976 with luminance."""

        uvy = Color('red').split_chromaticity('uv-1976')
        for v1, v2 in zip(uvy, [0.4507, 0.52289, 0.21264]):
            self.assertCompare(v1, v2)

    def test_new_xy_color(self):
        """Test new xy color."""

        c1 = Color('red')
        xyy = c1.split_chromaticity('xy-1931')
        c2 = Color.chromaticity('srgb', xyy, 'xy-1931')
        self.assertColorEqual(c1, c2)

    def test_new_uv_1960_color(self):
        """Test new uv 1960 color."""

        c1 = Color('red')
        uvy = c1.split_chromaticity('uv-1960')
        c2 = Color.chromaticity('srgb', uvy, 'uv-1960')
        self.assertColorEqual(c1, c2)

    def test_new_uv_1976_color(self):
        """Test new uv 1976 color."""

        c1 = Color('red')
        uvy = c1.split_chromaticity('uv-1976')
        c2 = Color.chromaticity('srgb', uvy, 'uv-1976')
        self.assertColorEqual(c1, c2)

    def test_chromaticity_bad_value(self):
        """Test bad chromaticity input."""

        with self.assertRaises(ValueError):
            Color.chromaticity('srgb', [0, 0], 'bad')

    def test_color_from_2D_chromaticity(self):
        """Test new color from 2D chromaticity."""

        c1 = Color('red')
        xy = c1.xy()
        c2 = Color.chromaticity('srgb', xy, 'xy-1931', scale=True, scale_space='srgb-linear')
        self.assertEqual(c2.to_string(fit=False), 'rgb(255 0 0)')
        [self.assertCompare(v1, v2) for v1, v2 in zip(c2.xy(), xy)]

    def test_uv_1960_to_xy(self):
        """Test `uv` 1960 to `xy`."""

        xy = Color('red').xy()
        uv = Color('red').uv('1960')

        for v1, v2 in zip(xy, cutil.uv_1960_to_xy(uv)):
            self.assertCompare(v1, v2)

    def test_uv_1976_to_xy(self):
        """Test `uv` 1960 to `xy`."""

        xy = Color('red').xy()
        uv = Color('red').uv('1976')

        for v1, v2 in zip(xy, cutil.uv_to_xy(uv)):
            self.assertCompare(v1, v2)

    def test_uv_1960_zero(self):
        """Test a case that should return zero."""

        u, v = cutil.xy_to_uv_1960((7.5, 1))
        self.assertCompare(u, 0)
        self.assertCompare(v, 0)

    def test_uv_1976_xy_zero(self):
        """Test a case that should return zero."""

        x, y = cutil.uv_to_xy((2 / 3, 1))
        self.assertCompare(x, 0)
        self.assertCompare(y, 0)

    def test_uv_1960_xy_zero(self):
        """Test a case that should return zero."""

        x, y = cutil.uv_1960_to_xy((2, 1))
        self.assertCompare(x, 0)
        self.assertCompare(y, 0)

    def test_bad_uv_value(self):
        """Test bad `uv` mode value."""

        with self.assertRaises(ValueError):
            Color('red').uv('bad')

    def test_chromaticity_no_xyz_value(self):
        """Test chromaticity does not accept XYZ values."""

        with self.assertRaises(ValueError):
            Color('red').chromaticity('srgb', [0.33, 0.66], 'xyz')

    def test_split_chromaticity_no_xyz_value(self):
        """Test that split chromaticity does not convert to XYZ."""

        with self.assertRaises(ValueError):
            Color('red').split_chromaticity('xyz')

    def test_convert_xyz_bad_coordinates(self):
        """Test chromaticity conversion with XYZ values."""

        with self.assertRaises(ValueError):
            Color.convert_chromaticity('xyz', 'xy-1931', [0, 0])

        with self.assertRaises(ValueError):
            Color.convert_chromaticity('xyz', 'xy-1931', [0, 0, 0, 0])

    def test_convert_chromaticity_bad_coordinates(self):
        """Test chromaticity conversion with XYZ values."""

        with self.assertRaises(ValueError):
            Color.convert_chromaticity('xy-1931', 'xyz', [0])

        with self.assertRaises(ValueError):
            Color.convert_chromaticity('xy-1931', 'xyz', [0, 0, 0, 0])

    def test_convert_chromaticity_no_convert(self):
        """Test case when no conversion is needed."""

        self.assertEqual(Color.convert_chromaticity('xy-1931', 'xy-1931', [0.33, 0.36, 1.0]), [0.33, 0.36, 1.0])

    def test_convert_chromaticity_no_convert_no_luminance(self):
        """Test case when no conversion is needed and no luminance is provided."""

        self.assertEqual(Color.convert_chromaticity('xy-1931', 'xy-1931', [0.33, 0.36]), [0.33, 0.36, 1.0])



"""
Model interpolation in a 3D space.

Interpolation can be done in any space, but displaying is currently only allowed in non polar spaces.
"""
import sys
import os
import matplotlib.pyplot as plt
import argparse

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from coloraide.spaces import Cylindrical  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='interp_3d_diagram', description='Demonstrate interpolation in 3D.')
    parser.add_argument('--space', '-s', default='srgb', help="Space to interpolate in.")
    parser.add_argument('--color', '-c', action='append', help="Color.")
    parser.add_argument('--display', '-d', help="Display space.")
    parser.add_argument('--method', '-m', default='linear', help="Interplation method to use: linear, bezier, etc.")
    parser.add_argument('--extrapolate', '-e', action='store_true', help='Extrapolate values.')
    parser.add_argument('--title', '-T', default='', help="Provide a title for the diagram.")
    parser.add_argument('--subtitle', '-t', default='', help="Provide a subtitle for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')
    args = parser.parse_args()

    if args.dark:
        edge = 'black'
        line = '#333333'
        plt.style.use('dark_background')
    else:
        edge = 'black'
        line = '#333333'
        plt.style.use('seaborn-v0_8-bright')

    # Create the color points
    x = []
    y = []
    z = []
    c = []
    colors = []
    display = None
    d = args.space if args.display is None else args.display
    for color in args.color:
        current = Color(color).convert(args.space)
        display = current.convert(d)
        x.append(display[0])
        y.append(display[1])
        z.append(display[2])
        c.append(current.convert('srgb').to_string(hex=True))
        colors.append(current)

    # We don't have logic to handle cylindrical color spaces.
    if isinstance(display._space, Cylindrical):
        raise ValueError('Displaying interpolation in a cylindrical color space is not currently supported.')

    # Setup figure and axis
    figure = plt.figure()
    plt.tight_layout()
    ax = plt.axes(
        projection='3d',
        xlabel=display._space.CHANNELS[0],
        ylabel=display._space.CHANNELS[1],
        zlabel=display._space.CHANNELS[2]
    )

    # Create title
    if not args.title:
        title = "{} Interpolation in '{}' space displayed in the '{}' space.".format(
            args.method.title(), args.space, d
        )
    else:
        title = args.title
    plt.suptitle(title)

    # Add axis to figure
    figure.add_axes(ax)

    # Interpolate between the entire range and optionally extrapolate
    i = Color.interpolate(
        colors,
        space=args.space,
        out_space=args.display,
        method=args.method,
        extrapolate=args.extrapolate
    )
    if not args.extrapolate:
        offset, factor = 0, 1
    else:
        offset, factor = 1, 3
    points = [i((r * factor / 1000) - offset)[:-1] for r in range(1001)]

    # Plot the interpolation line and plot the colors.
    xs, ys, zs = list(zip(*points))
    ax.plot(xs, ys, zs, color=line)
    ax.scatter(x, y, z, c=c, s=20 * 4, edgecolors=edge)

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Test Hunter Lab."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestHunterLab(util.ColorAssertsPyTest):
    """Test Hunter Lab."""

    COLORS = [
        ('red', 'color(--hunter-lab 46.113 82.672 28.408)'),
        ('orange', 'color(--hunter-lab 69.407 23.266 40.946)'),
        ('yellow', 'color(--hunter-lab 96.323 -21.054 55.869)'),
        ('green', 'color(--hunter-lab 39.291 -32.086 22.368)'),
        ('blue', 'color(--hunter-lab 26.869 75.478 -200.29)'),
        ('indigo', 'color(--hunter-lab 17.629 40.896 -62.916)'),
        ('violet', 'color(--hunter-lab 63.496 58.109 -40.51)'),
        ('white', 'color(--hunter-lab 100 0 0)'),
        ('gray', 'color(--hunter-lab 46.461 0 0)'),
        ('black', 'color(--hunter-lab 0 0 0)'),
        # Test color
        ('color(--hunter-lab 50 10 -10)', 'color(--hunter-lab 50 10 -10)'),
        ('color(--hunter-lab 50 10 -10 / 0.5)', 'color(--hunter-lab 50 10 -10 / 0.5)'),
        ('color(--hunter-lab 50% 50% -50% / 50%)', 'color(--hunter-lab 50 105 -105 / 0.5)'),
        ('color(--hunter-lab none none none / none)', 'color(--hunter-lab none none none / none)'),
        # Test percent ranges
        ('color(--hunter-lab 0% 0% 0%)', 'color(--hunter-lab 0 0 0)'),
        ('color(--hunter-lab 100% 100% 100%)', 'color(--hunter-lab 100 210 210)'),
        ('color(--hunter-lab -100% -100% -100%)', 'color(--hunter-lab -100 -210 -210)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('hunter-lab'), Color(color2))


class TestHunterLabSerialize(util.ColorAssertsPyTest):
    """Test Hunter Lab serialization."""

    COLORS = [
        # Test color
        ('color(--hunter-lab 75 10 -10 / 0.5)', {}, 'color(--hunter-lab 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--hunter-lab 75 10 -10)', {'alpha': True}, 'color(--hunter-lab 75 10 -10 / 1)'),
        ('color(--hunter-lab 75 10 -10 / 0.5)', {'alpha': False}, 'color(--hunter-lab 75 10 -10)'),
        # Test None
        ('color(--hunter-lab none 10 -10)', {}, 'color(--hunter-lab 0 10 -10)'),
        ('color(--hunter-lab none 10 -10)', {'none': True}, 'color(--hunter-lab none 10 -10)'),
        # Test Fit (not bound)
        ('color(--hunter-lab 120 10 -10)', {}, 'color(--hunter-lab 120 10 -10)'),
        ('color(--hunter-lab 120 10 -10)', {'fit': False}, 'color(--hunter-lab 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestHunterLabPoperties(util.ColorAsserts, unittest.TestCase):
    """Test Hunter Lab."""

    def test_l(self):
        """Test `l`."""

        c = Color('color(--hunter-lab 96.323 -21.06 55.728)')
        self.assertEqual(c['l'], 96.323)
        c['l'] = 0.2
        self.assertEqual(c['l'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--hunter-lab 96.323 -21.06 55.728)')
        self.assertEqual(c['a'], -21.06)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--hunter-lab 96.323 -21.06 55.728)')
        self.assertEqual(c['b'], 55.728)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hunter-lab 96.323 -21.06 55.728)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Test Jzazbz library."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestJzazbz(util.ColorAssertsPyTest):
    """Test Jzazbz."""

    COLORS = [
        ('red', 'color(--jzazbz 0.13438 0.11789 0.11188)'),
        ('orange', 'color(--jzazbz 0.16937 0.0312 0.12308)'),
        ('yellow', 'color(--jzazbz 0.2096 -0.02864 0.13479)'),
        ('green', 'color(--jzazbz 0.09203 -0.07454 0.07996)'),
        ('blue', 'color(--jzazbz 0.09577 -0.04085 -0.18585)'),
        ('indigo', 'color(--jzazbz 0.06146 0.03051 -0.09951)'),
        ('violet', 'color(--jzazbz 0.16771 0.06427 -0.05514)'),
        ('white', 'color(--jzazbz 0.22207 -0.00016 -0.00012)'),
        ('gray', 'color(--jzazbz 0.11827 -0.00012 -0.00008)'),
        ('black', 'color(--jzazbz 0 0 0)'),
        # Test color
        ('color(--jzazbz 0.5 0.1 -0.1)', 'color(--jzazbz 0.5 0.1 -0.1)'),
        ('color(--jzazbz 0.5 0.1 -0.1 / 0.5)', 'color(--jzazbz 0.5 0.1 -0.1 / 0.5)'),
        ('color(--jzazbz 50% 50% -50% / 50%)', 'color(--jzazbz 0.5 0.25 -0.25 / 0.5)'),
        ('color(--jzazbz none none none / none)', 'color(--jzazbz none none none / none)'),
        # Test percent ranges
        ('color(--jzazbz 0% 0% 0%)', 'color(--jzazbz 0 0 0)'),
        ('color(--jzazbz 100% 100% 100%)', 'color(--jzazbz 1 0.5 0.5)'),
        ('color(--jzazbz -100% -100% -100%)', 'color(--jzazbz -1 -0.5 -0.5)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('jzazbz'), Color(color2))


class TestJzazbzSerialize(util.ColorAssertsPyTest):
    """Test Jzazbz serialization."""

    COLORS = [
        # Test color
        ('color(--jzazbz 0.75 0.1 -0.1 / 0.5)', {}, 'color(--jzazbz 0.75 0.1 -0.1 / 0.5)'),
        # Test alpha
        ('color(--jzazbz 0.75 0.1 -0.1)', {'alpha': True}, 'color(--jzazbz 0.75 0.1 -0.1 / 1)'),
        ('color(--jzazbz 0.75 0.1 -0.1 / 0.5)', {'alpha': False}, 'color(--jzazbz 0.75 0.1 -0.1)'),
        # Test None
        ('color(--jzazbz none 0.1 -0.1)', {}, 'color(--jzazbz 0 0.1 -0.1)'),
        ('color(--jzazbz none 0.1 -0.1)', {'none': True}, 'color(--jzazbz none 0.1 -0.1)'),
        # Test Fit (not bound)
        ('color(--jzazbz 0.75 0.6 -0.1)', {}, 'color(--jzazbz 0.75 0.6 -0.1)'),
        ('color(--jzazbz 0.75 0.6 -0.1)', {'fit': False}, 'color(--jzazbz 0.75 0.6 -0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestJzazbzProperties(util.ColorAsserts, unittest.TestCase):
    """Test Jzazbz."""

    def test_jz(self):
        """Test `lightness`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c['jz'], 1)
        c['jz'] = 0.2
        self.assertEqual(c['jz'], 0.2)

    def test_az(self):
        """Test `az`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c['az'], 0.2)
        c['az'] = 0.1
        self.assertEqual(c['az'], 0.1)

    def test_bz(self):
        """Test `bz`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c['bz'], -0.3)
        c['bz'] = 0.1
        self.assertEqual(c['bz'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('jzazbz').is_achromatic(), True)
        self.assertEqual(Color('#222222').convert('jzazbz').set('jz', NaN).is_achromatic(), True)
        self.assertEqual(Color('jzazbz', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('jzazbz', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('jzazbz', [0, 0.3, -0.4]).is_achromatic(), True)
        self.assertEqual(Color('jzazbz', [NaN, 0, -0.3]).is_achromatic(), True)
        self.assertEqual(Color('jzazbz', [0.3, NaN, 0]).is_achromatic(), False)
        self.assertEqual(Color('jzazbz', [NaN, NaN, 0]).is_achromatic(), True)



"""Linear Pro Photo RGB color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector

RGB_TO_XYZ = [
    [0.7977604896723027, 0.13518583717574031, 0.0313493495815248],
    [0.2880711282292934, 0.7118432178101014, 8.565396060525902e-05],
    [0.0, 0.0, 0.8251046025104601]
]

XYZ_TO_RGB = [
    [1.3457989731028281, -0.2555801000799754, -0.05110628506753401],
    [-0.5446224939028347, 1.5082327413132781, 0.02053603239147973],
    [0.0, 0.0, 1.2119675456389454]
]


def lin_prophoto_to_xyz(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light prophoto-rgb values to CIE XYZ using  D50.D50.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    return alg.dot(RGB_TO_XYZ, rgb, dims=alg.D2_D1)


def xyz_to_lin_prophoto(xyz: Vector) -> Vector:
    """Convert XYZ to linear-light prophoto-rgb."""

    return alg.dot(XYZ_TO_RGB, xyz, dims=alg.D2_D1)


class ProPhotoRGBLinear(sRGB):
    """Linear Pro Photo RGB class."""

    BASE = "xyz-d50"
    NAME = "prophoto-rgb-linear"
    SERIALIZE = ('--prophoto-rgb-linear',)
    WHITE = WHITES['2deg']['D50']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Linear Pro Photo RGB."""

        return lin_prophoto_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Linear Pro Photo RGB."""

        return xyz_to_lin_prophoto(coords)



"""Test HSI."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
from coloraide import NaN
import pytest


class TestHSI(util.ColorAssertsPyTest):
    """Test HSI."""

    COLORS = [
        ('red', 'color(--hsi 0 1 0.33333)'),
        ('orange', 'color(--hsi 38.824 1 0.54902)'),
        ('yellow', 'color(--hsi 60 1 0.66667)'),
        ('green', 'color(--hsi 120 1 0.16732)'),
        ('blue', 'color(--hsi 240 1 0.33333)'),
        ('indigo', 'color(--hsi 274.62 1 0.26797)'),
        ('violet', 'color(--hsi 300 0.35644 0.79216)'),
        ('white', 'color(--hsi 0 0 1)'),
        ('gray', 'color(--hsi 0 0 0.50196)'),
        ('black', 'color(--hsi 0 0 0)'),
        # Test color
        ('color(--hsi 270 0.3 0.5)', 'color(--hsi 270 0.3 0.5)'),
        ('color(--hsi 270 0.3 0.5 / 0.5)', 'color(--hsi 270 0.3 0.5 / 0.5)'),
        ('color(--hsi 50% 50% 50% / 50%)', 'color(--hsi 180 0.5 0.5 / 0.5)'),
        ('color(--hsi none none none / none)', 'color(--hsi none none none / none)'),
        # Test percent ranges
        ('color(--hsi 0% 0% 0%)', 'color(--hsi 0 0 none)'),
        ('color(--hsi 100% 100% 100%)', 'color(--hsi 360 1 1 / 1)'),
        ('color(--hsi -100% -100% -100%)', 'color(--hsi -360 -1 -1 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_hsi_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('hsi'), Color(color2))


class TestHSISerialize(util.ColorAssertsPyTest):
    """Test HSI serialization."""

    COLORS = [
        # Test color
        ('color(--hsi 50 0.3 0.5 / 0.5)', {}, 'color(--hsi 50 0.3 0.5 / 0.5)'),
        # Test alpha
        ('color(--hsi 50 0.3 0.5)', {'alpha': True}, 'color(--hsi 50 0.3 0.5 / 1)'),
        ('color(--hsi 50 0.3 0.5 / 0.5)', {'alpha': False}, 'color(--hsi 50 0.3 0.5)'),
        # Test None
        ('color(--hsi 50 0.3 none)', {}, 'color(--hsi 50 0.3 0)'),
        ('color(--hsi 50 0.3 none)', {'none': True}, 'color(--hsi 50 0.3 none)'),
        # Test Fit (not bound)
        ('color(--hsi 50 1.1 0.5)', {}, 'color(--hsi 50 1 0.5)'),
        ('color(--hsi 50 1.1 0.5)', {'fit': False}, 'color(--hsi 50 1.1 0.5)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestHSIPoperties(util.ColorAsserts, unittest.TestCase):
    """Test HSI."""

    def test_names(self):
        """Test HSV-ish names."""

        self.assertEqual(Color('color(--hsi 60 1 0.66667)')._space.names(), ('h', 's', 'i'))

    def test_h(self):
        """Test `h`."""

        c = Color('color(--hsi 60 1 0.66667)')
        self.assertEqual(c['h'], 60)
        c['h'] = 0.2
        self.assertEqual(c['h'], 0.2)

    def test_s(self):
        """Test `s`."""

        c = Color('color(--hsi 60 1 0.66667)')
        self.assertEqual(c['s'], 1)
        c['s'] = 0.1
        self.assertEqual(c['s'], 0.1)

    def test_i(self):
        """Test `i`."""

        c = Color('color(--hsi 60 1 0.66667)')
        self.assertEqual(c['i'], 0.66667)
        c['i'] = 0.1
        self.assertEqual(c['i'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hsi 60 1 0.66667)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('hsi', [NaN, 0.5, 1], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--hsi none 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_sat(self):
        """Test minimum saturation."""

        c = Color('color(--hsi 270 0% 75% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_intensity(self):
        """Test minimum intensity."""

        c = Color('color(--hsi 270 20% 0% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_corner_case_null(self):
        """Test corner case that produces null."""

        c = Color('color(srgb -2 0 2)').convert('hsl')
        self.assertTrue(c.is_nan('hue'))



"""Test CAM16 UCS."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestCAM16CAM16(util.ColorAssertsPyTest):
    """Test CAM16."""

    COLORS = [
        ('red', 'color(--cam16 46.026 72.143 37.385)'),
        ('orange', 'color(--cam16 68.056 13.955 41.212)'),
        ('yellow', 'color(--cam16 94.682 -19.662 50.83)'),
        ('green', 'color(--cam16 33.976 -40.301 31.147)'),
        ('blue', 'color(--cam16 25.066 13.785 -60.901)'),
        ('indigo', 'color(--cam16 16.046 28.336 -32.712)'),
        ('violet', 'color(--cam16 63.507 41.066 -22.402)'),
        ('white', 'color(--cam16 100 -1.9463 -1.1026)'),
        ('gray', 'color(--cam16 43.042 -1.2764 -0.72317)'),
        ('black', 'color(--cam16 0 0 0)'),
        # Test color
        ('color(--cam16 50 10 -10)', 'color(--cam16 50 10 -10)'),
        ('color(--cam16 50 10 -10 / 0.5)', 'color(--cam16 50 10 -10 / 0.5)'),
        ('color(--cam16 50% 50% -50% / 50%)', 'color(--cam16 50 45 -45 / 0.5)'),
        ('color(--cam16 none none none / none)', 'color(--cam16 none none none / none)'),
        # Test percent ranges
        ('color(--cam16 0% 0% 0%)', 'color(--cam16 0 0 0)'),
        ('color(--cam16 100% 100% 100%)', 'color(--cam16 100 90 90)'),
        ('color(--cam16 -100% -100% -100%)', 'color(--cam16 -100 -90 -90)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cam16'), Color(color2))


class TestCAM16Serialize(util.ColorAssertsPyTest):
    """Test CAM16 serialization."""

    COLORS = [
        # Test color
        ('color(--cam16 75 10 -10 / 0.5)', {}, 'color(--cam16 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--cam16 75 10 -10)', {'alpha': True}, 'color(--cam16 75 10 -10 / 1)'),
        ('color(--cam16 75 10 -10 / 0.5)', {'alpha': False}, 'color(--cam16 75 10 -10)'),
        # Test None
        ('color(--cam16 none 10 -10)', {}, 'color(--cam16 0 10 -10)'),
        ('color(--cam16 none 10 -10)', {'none': True}, 'color(--cam16 none 10 -10)'),
        # Test Fit (not bound)
        ('color(--cam16 120 10 -10)', {}, 'color(--cam16 120 10 -10)'),
        ('color(--cam16 120 10 -10)', {'fit': False}, 'color(--cam16 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestCAM16Poperties(util.ColorAsserts, unittest.TestCase):
    """Test CAM16."""

    def test_j(self):
        """Test `j`."""

        c = Color('color(--cam16 0.51332 0.92781 1.076)')
        self.assertEqual(c['j'], 0.51332)
        c['j'] = 0.2
        self.assertEqual(c['j'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--cam16 0.51332 0.92781 1.076)')
        self.assertEqual(c['a'], 0.92781)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--cam16 0.51332 0.92781 1.076)')
        self.assertEqual(c['b'], 1.076)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cam16 0.51332 0.92781 1.076)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('cam16').is_achromatic(), True)
        self.assertEqual(Color('srgb', [0.000000001] * 3).convert('cam16').set('j', NaN).is_achromatic(), True)
        self.assertEqual(Color('cam16', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('cam16', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('cam16', [0, 3, -4]).is_achromatic(), True)
        self.assertEqual(Color('cam16', [NaN, 0, -3]).is_achromatic(), True)
        self.assertEqual(Color('cam16', [30, NaN, 0]).is_achromatic(), False)
        self.assertEqual(Color('cam16', [NaN, NaN, 0]).is_achromatic(), True)



"""Linear A98 RGB color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
from .. import algebra as alg
from ..types import Vector

RGB_TO_XYZ = [
    [0.5766690429101305, 0.1855582379065463, 0.1882286462349947],
    [0.29734497525053605, 0.6273635662554661, 0.07529145849399788],
    [0.02703136138641234, 0.07068885253582723, 0.9913375368376388]
]

XYZ_TO_RGB = [
    [2.0415879038107465, -0.5650069742788596, -0.34473135077832956],
    [-0.9692436362808795, 1.8759675015077202, 0.04155505740717558],
    [0.013444280632031142, -0.11836239223101837, 1.0151749943912054]
]


def lin_a98rgb_to_xyz(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light a98-rgb values to CIE XYZ using D50.D65.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    which has greater numerical precision than section 4.3.5.3 of
    https://www.adobe.com/digitalimag/pdfs/AdobeRGB1998.pdf
    """

    return alg.dot(RGB_TO_XYZ, rgb, dims=alg.D2_D1)


def xyz_to_lin_a98rgb(xyz: Vector) -> Vector:
    """Convert XYZ to linear-light a98-rgb."""

    return alg.dot(XYZ_TO_RGB, xyz, dims=alg.D2_D1)


class A98RGBLinear(sRGB):
    """Linear A98 RGB class."""

    BASE = "xyz-d65"
    NAME = "a98-rgb-linear"
    SERIALIZE = ('--a98-rgb-linear',)
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from A98 RGB."""

        return lin_a98rgb_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to A98 RGB."""

        return xyz_to_lin_a98rgb(coords)



"""Test Okhsl library."""
from coloraide import NaN
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestsOkhsl(util.ColorAssertsPyTest):
    """Test Okhsl."""

    COLORS = [
        ('red', 'color(--okhsl 29.234 1 0.56808)'),
        ('orange', 'color(--okhsl 70.67 1 0.75883)'),
        ('yellow', 'color(--okhsl 109.77 1 0.9627)'),
        ('green', 'color(--okhsl 142.5 1 0.44371)'),
        ('blue', 'color(--okhsl 264.05 1 0.36657)'),
        ('indigo', 'color(--okhsl 301.68 1 0.24043)'),
        ('violet', 'color(--okhsl 327.21 0.90625 0.7231)'),
        ('white', 'color(--okhsl 0 0 1)'),
        ('gray', 'color(--okhsl none 0 0.53571)'),
        ('black', 'color(--okhsl none 0 0)'),
        # Test color
        ('color(--okhsl 270 0.3 0.5)', 'color(--okhsl 270 0.3 0.5)'),
        ('color(--okhsl 270 0.3 0.5 / 0.5)', 'color(--okhsl 270 0.3 0.5 / 0.5)'),
        ('color(--okhsl 50% 50% 50% / 50%)', 'color(--okhsl 180 0.5 0.5 / 0.5)'),
        ('color(--okhsl none none none / none)', 'color(--okhsl none none none / none)'),
        # Test percent ranges
        ('color(--okhsl 0% 0% 0%)', 'color(--okhsl 0 0 none)'),
        ('color(--okhsl 100% 100% 100%)', 'color(--okhsl 360 1 1 / 1)'),
        ('color(--okhsl -100% -100% -100%)', 'color(--okhsl -360 -1 -1 / 1)'),
        # Space specific tests to ensure complete coverage in algorithm
        ('color(--oklab 0.5 0.1 0)', 'color(--okhsl 0 0.49656 0.42114)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('okhsl'), Color(color2))


class TestOkhslSerialize(util.ColorAssertsPyTest):
    """Test Okhsl serialization."""

    COLORS = [
        # Test color
        ('color(--okhsl 50 0.3 0.5 / 0.5)', {}, 'color(--okhsl 50 0.3 0.5 / 0.5)'),
        # Test alpha
        ('color(--okhsl 50 0.3 0.5)', {'alpha': True}, 'color(--okhsl 50 0.3 0.5 / 1)'),
        ('color(--okhsl 50 0.3 0.5 / 0.5)', {'alpha': False}, 'color(--okhsl 50 0.3 0.5)'),
        # Test None
        ('color(--okhsl 50 0.3 none)', {}, 'color(--okhsl 50 0.3 0)'),
        ('color(--okhsl 50 0.3 none)', {'none': True}, 'color(--okhsl 50 0.3 none)'),
        # Test Fit (not bound)
        ('color(--okhsl 50 1.1 0.5)', {}, 'color(--okhsl 50 1 0.5)'),
        ('color(--okhsl 50 1.1 0.5)', {'fit': False}, 'color(--okhsl 50 1.1 0.5)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestOkhslProperties(util.ColorAsserts, unittest.TestCase):
    """Test Okhsl."""

    def test_names(self):
        """Test HSL-ish names."""

        self.assertEqual(Color('color(--okhsl 120 50% 90% / 1)')._space.names(), ('h', 's', 'l'))

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c['hue'], 120)
        c['hue'] = 110
        self.assertEqual(c['hue'], 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c['saturation'], 0.5)
        c['saturation'] = 0.6
        self.assertEqual(c['saturation'], 0.6)

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c['lightness'], 0.9)
        c['lightness'] = 0.8
        self.assertEqual(c['lightness'], 0.8)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('okhsl', [NaN, 0.5, 1], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--okhsl none 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_sat(self):
        """Test minimum saturation."""

        c = Color('color(--okhsl 270 0% 0.75 / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_max_light(self):
        """Test maximum lightness."""

        c = Color('color(--okhsl 270 20% 100% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_light(self):
        """Test minimum lightness."""

        c = Color('color(--okhsl 270 20% 0% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))



"""
HSI class.

https://en.wikipedia.org/wiki/HSL_and_HSV#Saturation
"""
from __future__ import annotations
from .hsv import HSV
from ..cat import WHITES
from ..channels import Channel, FLG_ANGLE
from .. import algebra as alg
from .. import util
from ..types import Vector


def srgb_to_hsi(rgb: Vector) -> Vector:
    """Convert sRGB to HSI."""

    r, g, b = rgb
    h = 0.0
    s = 0.0
    mx = max(rgb)
    mn = min(rgb)
    i = sum(rgb) * 1 / 3
    s = 0 if i == 0.0 else 1 - (mn / i)
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        h *= 60.0

    return [util.constrain_hue(h), s, i]


def hsi_to_srgb(hsi: Vector) -> Vector:
    """HSI to RGB."""

    h, s, i = hsi
    h = util.constrain_hue(h) / 60
    z = 1 - abs(h % 2 - 1)
    c = (3 * i * s) / (1 + z)
    x = c * z

    if alg.is_nan(h):  # pragma: no cover
        # In our current setup, this will not occur. If colors are naturally achromatic,
        # they will resolve to zeros automatically even without this check. This case
        # would be a shortcut normally.
        #
        # Unnatural cases, such as explicitly setting of hue to undefined, could cause this,
        # but the conversion pipeline converts all undefined values to zero. We'd have to
        # encounter a natural case due to conversion to or from HSI mid pipeline to trigger
        # this, and we are not currently in a position where that would occur with sRGB being
        # the only pass-through.
        rgb = [0.0] * 3
    elif 0 <= h <= 1:
        rgb = [c, x, 0]
    elif 1 <= h <= 2:
        rgb = [x, c, 0]
    elif 2 <= h <= 3:
        rgb = [0, c, x]
    elif 3 <= h <= 4:
        rgb = [0, x, c]
    elif 4 <= h <= 5:
        rgb = [x, 0, c]
    else:
        rgb = [c, 0, x]
    m = i * (1 - s)

    return [chan + m for chan in rgb]


class HSI(HSV):
    """HSI class."""

    BASE = "srgb"
    NAME = "hsi"
    SERIALIZE = ("--hsi",)
    CHANNELS = (
        Channel("h", 0.0, 360.0, bound=True, flags=FLG_ANGLE),
        Channel("s", 0.0, 1.0, bound=True),
        Channel("i", 0.0, 1.0, bound=True)
    )
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "intensity": "i"
    }
    WHITE = WHITES['2deg']['D65']
    GAMUT_CHECK = "srgb"

    def to_base(self, coords: Vector) -> Vector:
        """To sRGB from HSI."""

        return hsi_to_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From sRGB to HSI."""

        return srgb_to_hsi(coords)



"""Test IPT."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestIPT(util.ColorAssertsPyTest):
    """Test IPT."""

    COLORS = [
        ('red', 'color(--ipt 0.45616 0.62091 0.44282)'),
        ('orange', 'color(--ipt 0.64877 0.18903 0.5303)'),
        ('yellow', 'color(--ipt 0.85655 -0.1083 0.65147)'),
        ('green', 'color(--ipt 0.3933 -0.23454 0.27459)'),
        ('blue', 'color(--ipt 0.44429 -0.2372 -0.74849)'),
        ('indigo', 'color(--ipt 0.28317 0.09827 -0.31015)'),
        ('violet', 'color(--ipt 0.72886 0.31225 -0.24012)'),
        ('white', 'color(--ipt 0.99999 0.00007 -0.00004)'),
        ('gray', 'color(--ipt 0.51724 0.00003 -0.00002)'),
        ('black', 'color(--ipt 0 0 0)'),
        # Test color
        ('color(--ipt 0.5 0.1 -0.1)', 'color(--ipt 0.5 0.1 -0.1)'),
        ('color(--ipt 0.5 0.1 -0.1 / 0.5)', 'color(--ipt 0.5 0.1 -0.1 / 0.5)'),
        ('color(--ipt 50% 50% -50% / 50%)', 'color(--ipt 0.5 0.5 -0.5 / 0.5)'),
        ('color(--ipt none none none / none)', 'color(--ipt none none none / none)'),
        # Test percent ranges
        ('color(--ipt 0% 0% 0%)', 'color(--ipt 0 0 0)'),
        ('color(--ipt 100% 100% 100%)', 'color(--ipt 1 1 1)'),
        ('color(--ipt -100% -100% -100%)', 'color(--ipt -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('ipt'), Color(color2))


class TestIPTSerialize(util.ColorAssertsPyTest):
    """Test IPT serialization."""

    COLORS = [
        # Test color
        ('color(--ipt 0.75 0.1 -0.1 / 0.5)', {}, 'color(--ipt 0.75 0.1 -0.1 / 0.5)'),
        # Test alpha
        ('color(--ipt 0.75 0.1 -0.1)', {'alpha': True}, 'color(--ipt 0.75 0.1 -0.1 / 1)'),
        ('color(--ipt 0.75 0.1 -0.1 / 0.5)', {'alpha': False}, 'color(--ipt 0.75 0.1 -0.1)'),
        # Test None
        ('color(--ipt none 0.1 -0.1)', {}, 'color(--ipt 0 0.1 -0.1)'),
        ('color(--ipt none 0.1 -0.1)', {'none': True}, 'color(--ipt none 0.1 -0.1)'),
        # Test Fit (not bound)
        ('color(--ipt 0.75 1.2 -0.1)', {}, 'color(--ipt 0.75 1.2 -0.1)'),
        ('color(--ipt 0.75 1.2 -0.1)', {'fit': False}, 'color(--ipt 0.75 1.2 -0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestIPTPoperties(util.ColorAsserts, unittest.TestCase):
    """Test IPT."""

    def test_i(self):
        """Test `i`."""

        c = Color('color(--ipt 0.85654 -0.10831 0.65148)')
        self.assertEqual(c['i'], 0.85654)
        c['i'] = 0.2
        self.assertEqual(c['i'], 0.2)

    def test_p(self):
        """Test `p`."""

        c = Color('color(--ipt 0.85654 -0.10831 0.65148)')
        self.assertEqual(c['p'], -0.10831)
        c['p'] = 0.1
        self.assertEqual(c['p'], 0.1)

    def test_t(self):
        """Test `t`."""

        c = Color('color(--ipt 0.85654 -0.10831 0.65148)')
        self.assertEqual(c['t'], 0.65148)
        c['t'] = 0.1
        self.assertEqual(c['t'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--ipt 0.85654 -0.10831 0.65148)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#464646').convert('ipt').is_achromatic(), True)
        self.assertEqual(Color('ipt', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('ipt', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('ipt', [0, 0.3, -0.4]).is_achromatic(), False)
        self.assertEqual(Color('ipt', [NaN, 0, -0.3]).is_achromatic(), False)
        self.assertEqual(Color('ipt', [0.3, NaN, 0]).is_achromatic(), False)
        self.assertEqual(Color('ipt', [NaN, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('ipt', [-0.5, NaN, NaN]).is_achromatic(), True)



"""
CIE 1960 UCS color class.

http://en.wikipedia.org/wiki/CIE_1960_color_space#Relation_to_CIE_XYZ
"""
from __future__ import annotations
from ..spaces import Space
from ..channels import Channel
from ..cat import WHITES
from ..types import Vector


def xyz_to_ucs(xyz: Vector) -> Vector:
    """Translate XYZ to 1960 UCS."""

    x, y, z = xyz
    return [(2 / 3) * x, y, (-x + 3 * y + z) * 0.5]


def ucs_to_xyz(ucs: Vector) -> Vector:
    """Translate 1960 UCS to XYZ."""

    u, v, w = ucs
    return [(3 / 2) * u, v, (3 / 2) * u - 3 * v + 2 * w]


class UCS(Space):
    """The 1960 UCS class."""

    BASE = "xyz-d65"
    NAME = "ucs"
    SERIALIZE = ("--ucs",)
    CHANNELS = (
        Channel("u", 0.0, 1.0),
        Channel("v", 0.0, 1.0),
        Channel("w", 0.0, 1.0)
    )
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return ucs_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return xyz_to_ucs(coords)



"""Generate all 3D models."""
import sys
import os
import argparse

sys.path.insert(0, os.getcwd())

import tools.gamut_3d_plotly as diagrams  # noqa: E402

LOCATION = 'docs/src/markdown/images'
TEMPLATE = 'sRGB Gamut Plotted in {} Color Space'


def plot_model(name, title, filename, gamut='srgb', elev=45, azim=-60.0):
    """Generate the models."""

    print('===> Generating {} model...'.format(name))
    fig = diagrams.plot_gamut_in_space(
        name,
        gamut=gamut,
        title=title,
        resolution=200,
        size=(800, 800),
        camera={'a': azim, 'e': elev, 'r': 2.5}
    )

    with open(os.path.join(LOCATION, filename), 'wb') as f:
        f.write(fig.to_image(format='png'))
    del fig
    print('[complete]')


models = {
    'srgb': {'title': 'sRGB Color Space', 'filename': 'srgb-3d.png'},
    'hsl': {'title': 'HSL Color Space', 'filename': 'hsl-3d.png'},
    'hsi': {'title': 'HSI Color Space', 'filename': 'hsi-3d.png'},
    'hsv': {'title': 'HSV Color Space', 'filename': 'hsv-3d.png'},
    'hwb': {'title': 'HWB Color Space', 'filename': 'hwb-3d.png', 'elev': 210},
    'okhsl': {'title': 'Okhsl Color Space', 'filename': 'okhsl-3d.png'},
    'okhsv': {'title': 'Okhsv Color Space', 'filename': 'okhsv-3d.png'},
    'hsluv': {'title': 'HSLuv Color Space', 'filename': 'hsluv-3d.png'},
    'hpluv': {'title': 'HPLuv Color Space', 'filename': 'hpluv-3d.png'},
    'xyz-d50': {'title': TEMPLATE.format('XYZ D50'), 'filename': 'xyz-d50-3d.png'},
    'xyz-d65': {'title': TEMPLATE.format('XYZ D65'), 'filename': 'xyz-d65-3d.png'},
    'xyy': {'title': TEMPLATE.format('xyY'), 'filename': 'xyy-3d.png'},
    'lab': {'title': TEMPLATE.format('CIELab D50'), 'filename': 'lab-3d.png'},
    'lab-d65': {'title': TEMPLATE.format('CIELab D65'), 'filename': 'lab-d65-3d.png'},
    'lch': {'title': TEMPLATE.format('CIELCh D50'), 'filename': 'lch-3d.png', 'azim': 300},
    'lch-d65': {'title': TEMPLATE.format('CIELCh D65'), 'filename': 'lch-d65-3d.png', 'azim': 300},
    'oklab': {'title': TEMPLATE.format('Oklab'), 'filename': 'oklab-3d.png'},
    'oklch': {'title': TEMPLATE.format('OkLCh'), 'filename': 'oklch-3d.png', 'azim': 300},
    'luv': {'title': TEMPLATE.format('CIELuv'), 'filename': 'luv-3d.png'},
    'lchuv': {'title': TEMPLATE.format('LChuv'), 'filename': 'lchuv-3d.png', 'azim': 300},
    'din99o': {'title': TEMPLATE.format('DIN99o'), 'filename': 'din99o-3d.png'},
    'lch99o': {'title': TEMPLATE.format('DIN99o LCh'), 'filename': 'lch99o-3d.png', 'azim': 300},
    'jzazbz': {'title': TEMPLATE.format('Jzazbz'), 'filename': 'jzazbz-3d.png'},
    'jzczhz': {'title': TEMPLATE.format('JzCzhz'), 'filename': 'jzczhz-3d.png', 'azim': 300},
    'rlab': {'title': TEMPLATE.format('RLAB'), 'filename': 'rlab-3d.png'},
    'hunter-lab': {'title': TEMPLATE.format('Hunter Lab'), 'filename': 'hunter-lab-3d.png', 'elev': 5, 'azim': -30},
    'ipt': {'title': TEMPLATE.format('IPT'), 'filename': 'ipt-3d.png'},
    'ictcp': {'title': TEMPLATE.format('ICtCp'), 'filename': 'ictcp-3d.png', 'azim': 60},
    'igpgtg': {'title': TEMPLATE.format('IgPgTg'), 'filename': 'igpgtg-3d.png'},
    'cam16': {'title': TEMPLATE.format('CAM16'), 'filename': 'cam16-3d.png'},
    'cam16-ucs': {'title': TEMPLATE.format('CAM16 UCS'), 'filename': 'cam16-ucs-3d.png'},
    'cam16-scd': {'title': TEMPLATE.format('CAM16 SCD'), 'filename': 'cam16-scd-3d.png'},
    'cam16-lcd': {'title': TEMPLATE.format('CAM16 LCD'), 'filename': 'cam16-lcd-3d.png'},
    'cam16-jmh': {'title': TEMPLATE.format('CAM16 JMh'), 'filename': 'cam16-jmh-3d.png'},
    'hct': {'title': TEMPLATE.format('HCT'), 'filename': 'hct-3d.png'},
    'cmy': {'title': TEMPLATE.format('CMY'), 'filename': 'cmy-3d.png'},
    'orgb': {'title': TEMPLATE.format('oRGB'), 'filename': 'orgb-3d.png', 'elev': 72},
    'xyb': {'title': TEMPLATE.format('XYB'), 'filename': 'xyb-3d.png', 'azim': 45},
    'ucs': {'title': TEMPLATE.format('UCS'), 'filename': 'ucs-3d.png', 'azim': 60, 'elev': 10},
}


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='gen_3d_models', description='Generate 3D models of color spaces.')
    parser.add_argument('--space', '-s', action='append', help="Color space.")
    args = parser.parse_args()

    if args.space:
        for space in args.space:
            kwargs = models.get(space)
            if kwargs is not None:
                plot_model(space, **kwargs)
    else:
        for space, kwargs in models.items():
            plot_model(space, **kwargs)


if __name__ == "__main__":
    sys.exit(main())



"""Test compositing."""
import unittest
from coloraide import Color, NaN
from . import util


class TestCompositing(util.ColorAsserts, unittest.TestCase):
    """Test compositing."""

    def test_disable_compose(self):
        """Test that we can disable either blend or alpha compositing."""

        c1 = Color('#07c7ed').set('alpha', 0.5).compose('#fc3d99', blend='multiply', operator=False, space="srgb")
        c2 = c1.compose('#fc3d99', blend=False, space="srgb")
        self.assertColorEqual(
            Color('#07c7ed').set('alpha', 0.5).compose('#fc3d99', blend='multiply', space="srgb"),
            c2
        )
        self.assertColorEqual(
            Color('#07c7ed').set('alpha', 0.5).compose('#fc3d99', blend=False, operator=False, space="srgb"),
            Color('#07c7ed').set('alpha', 0.5)
        )

    def test_compose_no_operator(self):
        """Test compose with no operator."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        c3 = c1.compose(c2)
        c4 = c1.compose(c2, operator='source-over')
        self.assertEqual(c3, c4)

    def test_compose(self):
        """Test compose logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        c3 = c1.compose(c2)
        self.assertTrue(c1 is not c3)
        self.assertEqual(c1.compose(c2), Color('color(srgb 0.5 0.5 0.5)'))

    def test_compose_dict(self):
        """Test compositing with a mapping."""

        c1 = Color('blue').set('alpha', 0.5)
        self.assertEqual(
            c1.compose("yellow"),
            c1.compose({"space": "srgb", "coords": [1, 1, 0]})
        )

    def test_compose_blend_multi(self):
        """Test compose blend with multiple colors."""

        self.assertColorEqual(
            Color('#07c7ed').compose(['#fc3d99', '#f5d311'], blend='multiply', space="srgb"),
            Color('rgb(6.6464 39.39 9.48)')
        )

    def test_compose_alpha_multi(self):
        """Test compose alpha compositing with multiple colors."""

        self.assertColorEqual(
            Color('#07c7ed').set('alpha', 0.5).compose(
                [Color('#fc3d99').set('alpha', 0.5), Color('#f5d311').set('alpha', 0.5), 'white'],
                blend='normal',
                space="srgb"
            ),
            Color('rgb(129 173 190.75)')
        )

    def test_compose_empty_list(self):
        """Test compose with empty list."""

        self.assertColorEqual(Color('green').compose([]), Color('green'))

    def test_compose_bad_operator(self):
        """Test compose bad operator."""

        with self.assertRaises(ValueError):
            Color('red').compose('blue', operator='bad')

    def test_compose_nan(self):
        """Test compose with `NaN` values."""

        self.assertColorEqual(
            Color('srgb', [NaN, 0.75, 0.75], 0.5).compose(Color('srgb', [1, 0.25, 0.25])),
            Color('rgb(127.5 127.5 127.5)')
        )
        self.assertColorEqual(
            Color('srgb', [NaN, 0.75, 0.75], 0.5).compose(Color('srgb', [NaN, 0.25, 0.25])),
            Color('rgb(0 127.5 127.5)')
        )
        self.assertColorEqual(
            Color('srgb', [0.2, 0.75, 0.75], 0.5).compose(Color('srgb', [NaN, 0.25, 0.25])),
            Color('rgb(25.5 127.5 127.5)')
        )

    def test_compose_bad_space(self):
        """Test compose logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        with self.assertRaises(KeyError):
            c1.compose(c2, space="bad")

    def test_compose_no_alpha(self):
        """Test compose logic when color has no alpha."""

        c1 = Color('blue')
        c2 = Color('yellow')
        c3 = c1.compose(c2)
        self.assertTrue(c1 is not c3)
        self.assertEqual(c1.compose(c2), c1)

    def test_compose_nan_alpha(self):
        """Test compose logic with alpha as `NaN`."""

        c1 = Color('blue').set('alpha', NaN)
        c2 = Color('yellow')
        c3 = c1.compose(c2)
        self.assertTrue(c1 is not c3)
        self.assertEqual(c3, Color('color(srgb 1 1 0 / 1)'))

    def test_compose_in_place(self):
        """Test compose logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        c3 = c1.compose(c2, in_place=True)
        self.assertTrue(c1 is c3)
        self.assertEqual(c1, Color('color(srgb 0.5 0.5 0.5)'))

    def test_compose_cyl(self):
        """Test compose logic."""

        c1 = Color('blue').set('alpha', 0.5)
        c2 = Color('yellow')
        with self.assertRaises(ValueError):
            c1.compose(c2, space="hsl")



"""XYZ D65 class."""
from __future__ import annotations
import math
from ..spaces import Space, RGBish
from ..cat import WHITES
from ..channels import Channel
from .. import util
from .. import algebra as alg
from ..types import Vector


class XYZD65(RGBish, Space):
    """XYZ D65 class."""

    BASE = "xyz-d65"
    NAME = "xyz-d65"
    SERIALIZE = ("xyz-d65", 'xyz')  # type: tuple[str, ...]
    CHANNELS = (
        Channel("x", 0.0, 1.0),
        Channel("y", 0.0, 1.0),
        Channel("z", 0.0, 1.0)
    )
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Is achromatic."""

        for x in alg.vcross(coords, util.xy_to_xyz(self.white())):
            if not math.isclose(0.0, x, abs_tol=1e-5):
                return False
        return True

    def to_base(self, coords: Vector) -> Vector:
        """
        To XYZ (no change).

        Any needed chromatic adaptation is handled in the parent Color object.
        """

        return coords

    def from_base(self, coords: Vector) -> Vector:
        """
        From XYZ (no change).

        Any needed chromatic adaptation is handled in the parent Color object.
        """

        return coords



"""Color harmonies."""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from . import algebra as alg
from .spaces import Cylindrical
from .cat import WHITES
from . import util
from typing import TYPE_CHECKING, Any  # noqa: F401

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color

WHITE = util.xy_to_xyz(WHITES['2deg']['D65'])
BLACK = [0, 0, 0]


def adjust_hue(hue: float, deg: float) -> float:
    """Adjust hue by the given degree."""

    return hue + deg


class Harmony(metaclass=ABCMeta):
    """Color harmony."""

    @abstractmethod
    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""


class Monochromatic(Harmony):
    """
    Monochromatic harmony.

    Take a given color and create both tints and shades such that we have `RANGE` total steps ranging
    from black -> color -> white. Normally, we will throw away pure black, pure white, and the duplicate
    target color (as we interpolate with it on both sides) leaving us with RANGE - 3 colors to extract
    the target `STEPS` from. The one exception is when targeting an achromatic color, and in that case,
    we only throw away the duplicate color (though if a color is close enough to white or black, white
    or black may not be included simply because we cannot get a reasonable step that includes it).

    Once we have our `RANGE`, we can extract a total of `STEPS` colors with the target color at the center
    (when possible). If the target color is too close to the either the minimum or maximum color step,
    there may not be enough tints or shades on one side, so the result may have to draw heavier on the
    side that has more plentiful tints or shades which would cause the target color to shift from the center.

    The current `RANGE` was chosen as 12 as it seems to to provide OK contrast in most cases for the monochromatic
    colors. The one exception is with a target color of black or very near black which may return at least one color
    with very low contrast to black. Generally, extremely dark colors do not make a good target for color harmonies,
    but it should be noted that OkLCh's lightness tends to the more darker side. The poor contrast may be
    less with other color spaces.
    """

    DELTA_E = '2000'
    RANGE = 12
    STEPS = 5

    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""

        color1 = color.convert(space, norm=False).normalize()

        if not isinstance(color1._space, Cylindrical):
            raise ValueError('Color space must be cylindrical')

        # Trim off black and white unless the color is achromatic,
        # But always trim duplicate target color from left side.
        if not color1.is_nan('hue'):
            ltrim, rtrim = slice(1, -1, None), slice(None, -1, None)
        else:
            ltrim, rtrim = slice(None, -1, None), slice(None, None, None)

        # Create black and white so we can generate tints and shades
        # Ensure hue and alpha is masked so we don't interpolate them.
        w = color.new('xyz-d65', WHITE, alg.NaN)
        w.convert(space, in_place=True, norm=False).mask(['hue', 'alpha'], in_place=True)
        b = color.new('xyz-d65', BLACK, alg.NaN)
        b.convert(space, in_place=True, norm=False).mask(['hue', 'alpha'], in_place=True)

        # Calculate how many tints and shades we need to generate
        db = b.delta_e(color1, method=self.DELTA_E)
        dw = w.delta_e(color1, method=self.DELTA_E)
        steps_w = int(alg.round_half_up((dw / (db + dw)) * self.RANGE))
        steps_b = self.RANGE - steps_w

        kwargs = {
            'space': space,
            'method': 'linear',
            'out_space': space
        }  # type: dict[str, Any]

        # Very close to black or is black, no need to interpolate from black to current color
        if steps_b <= 1:
            left = []
            if steps_b == 1:
                left.extend(color1.steps([b, color1], steps=steps_b, **kwargs))
            steps = min(self.RANGE - (1 + steps_b), steps_w)
            right = color1.steps([color1, w], steps=steps, **kwargs)[rtrim]

        # Very close to white or is white, no need to interpolate from current color to white
        elif steps_w <= 1:
            right = []
            if steps_w == 1:
                right.extend(color1.steps([color1, w], steps=steps_w, **kwargs))
            steps = min(self.RANGE - (1 + steps_w), steps_b)
            right.insert(0, color1.clone())
            left = color1.steps([b, color1], steps=steps, **kwargs)[ltrim]

        # Anything else in between
        else:
            left = color1.steps([b, color1], steps=steps_b, **kwargs)[ltrim]
            right = color1.steps([color1, w], steps=steps_w, **kwargs)[rtrim]

        # Extract a subset of the results
        len_l = len(left)
        len_r = len(right)
        l = int(self.STEPS // 2)
        r = l + (1 if self.STEPS % 2 else 0)
        if len_r < r:
            return left[-self.STEPS + len_r:] + right
        elif len_l < l:
            return left + right[:self.STEPS - len_l]
        return left[-l:] + right[:r]


class Geometric(Harmony):
    """Geometrically space the colors."""

    COUNT = 0

    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""

        color1 = color.convert(space, norm=False).normalize()

        if not isinstance(color1._space, Cylindrical):
            raise ValueError('Color space must be cylindrical')

        name = color1._space.hue_name()

        degree = current = 360.0 / self.COUNT
        colors = []
        for _ in range(self.COUNT - 1):
            colors.append(
                color1.clone().set(name, lambda x, value=current: adjust_hue(x, value))
            )
            current += degree
        colors.insert(0, color1)
        return colors


class Complementary(Geometric):
    """Complementary colors."""

    COUNT = 2


class Triadic(Geometric):
    """Triadic colors."""

    COUNT = 3


class TetradicSquare(Geometric):
    """Tetradic (square)."""

    COUNT = 4


class SplitComplementary(Harmony):
    """Split Complementary colors."""

    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""

        color1 = color.convert(space, norm=False).normalize()

        if not isinstance(color1._space, Cylindrical):
            raise ValueError('Color space must be cylindrical')

        name = color1._space.hue_name()

        color2 = color1.clone().set(name, lambda x: adjust_hue(x, 210))
        color3 = color1.clone().set(name, lambda x: adjust_hue(x, -210))
        return [color1, color2, color3]


class Analogous(Harmony):
    """Analogous colors."""

    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""

        color1 = color.convert(space, norm=False).normalize()

        if not isinstance(color1._space, Cylindrical):
            raise ValueError('Color space must be cylindrical')

        name = color1._space.hue_name()

        color2 = color1.clone().set(name, lambda x: adjust_hue(x, 30))
        color3 = color1.clone().set(name, lambda x: adjust_hue(x, -30))
        return [color1, color2, color3]


class TetradicRect(Harmony):
    """Tetradic (rectangular) colors."""

    def harmonize(self, color: Color, space: str) -> list[Color]:
        """Get color harmonies."""

        color1 = color.convert(space, norm=False).normalize()

        if not isinstance(color1._space, Cylindrical):
            raise ValueError('Color space must be cylindrical')

        name = color1._space.hue_name()

        color2 = color1.clone().set(name, lambda x: adjust_hue(x, 30))
        color3 = color1.clone().set(name, lambda x: adjust_hue(x, 180))
        color4 = color1.clone().set(name, lambda x: adjust_hue(x, 210))
        return [color1, color2, color3, color4]


SUPPORTED = {
    'complement': Complementary(),
    'split': SplitComplementary(),
    'triad': Triadic(),
    'square': TetradicSquare(),
    'rectangle': TetradicRect(),
    'analogous': Analogous(),
    'mono': Monochromatic()
}  # type: dict[str, Harmony]


def harmonize(color: Color, name: str, space: str) -> list[Color]:
    """Get specified color harmonies."""

    h = SUPPORTED.get(name)
    if not h:
        raise ValueError("The color harmony '{}' cannot be found".format(name))

    return h.harmonize(color, space)



"""Average colors together."""
from __future__ import annotations
import math
from .types import ColorInput
from typing import Iterable, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color


def average(create: type[Color], colors: Iterable[ColorInput], space: str, premultiplied: bool = True) -> Color:
    """Average a list of colors together."""

    obj = create(space, [])

    # Get channel information
    cs = obj.CS_MAP[space]
    hue_index = cs.hue_index() if hasattr(cs, 'hue_index') else -1
    channels = cs.channels
    chan_count = len(channels)
    alpha_index = chan_count - 1
    sums = [0.0] * chan_count
    totals = [0.0] * chan_count
    sin = 0.0
    cos = 0.0

    # Sum channel values
    i = -1
    for c in colors:
        coords = obj.update(c)[:]
        alpha = coords[-1]
        if math.isnan(alpha):
            alpha = 1.0
        i = 0
        for coord in coords:
            if not math.isnan(coord):
                totals[i] += 1
                if i == hue_index:
                    rad = math.radians(coord)
                    sin += math.sin(rad)
                    cos += math.cos(rad)
                else:
                    sums[i] += (coord * alpha) if premultiplied and i != alpha_index else coord
            i += 1

    if i == -1:
        raise ValueError('At least one color must be provided in order to average colors')

    # Get the mean
    alpha = sums[-1]
    alpha_t = totals[-1]
    sums[-1] = float('nan') if not alpha_t else alpha / alpha_t
    alpha = sums[-1]
    if math.isnan(alpha) or alpha in (0.0, 1.0):
        alpha = 1.0
    for i in range(chan_count - 1):
        total = totals[i]
        if not total:
            sums[i] = float('nan')
        elif i == hue_index:
            sums[i] = math.degrees(math.atan2(sin / total, cos / total))
        else:
            sums[i] /= total * alpha if premultiplied else total

    # Return the color
    return obj.update(space, sums[:-1], sums[-1])



"""LChuv class."""
from __future__ import annotations
from ..spaces import Space
from ..cat import WHITES
from ..channels import Channel, FLG_ANGLE
from .lch import LCh, ACHROMATIC_THRESHOLD
from ..types import Vector


class LChuv(LCh, Space):
    """LChuv class."""

    BASE = "luv"
    NAME = "lchuv"
    SERIALIZE = ("--lchuv",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("c", 0.0, 220.0, limit=(0.0, None)),
        Channel("h", 0.0, 360.0, flags=FLG_ANGLE)
    )

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return coords[0] == 0.0 or coords[1] < ACHROMATIC_THRESHOLD



"""
Handle pointer gamut.

Data used for the gamut: https://www.rit-mcsl.org/UsefulData/PointerData.xls.
"""
from __future__ import annotations
import math
import bisect
from ..spaces.lab import xyz_to_lab, lab_to_xyz
from ..spaces.lch import lab_to_lch, lch_to_lab
from .. import algebra as alg
from .. import util
from ..types import Vector
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color

# White point C as defined in the Pointer data spreadsheet
XYZ_W = (98.0722647623506, 100.0, 118.225418982695)
WHITE_POINT_SC = tuple(util.xyz_to_xyY(XYZ_W)[:-1])  # type: tuple[float, float]  # type: ignore[assignment]
# Rows: hue 0 - 350 at steps of 10
# Columns: lightness 15 - 90 at steps of 5
LCH_L = list(range(15, 91, 5))
LCH_H = list(range(0, 351, 10))
LCH_POINTER = [
    [10, 30, 43, 56, 68, 77, 79, 77, 72, 65, 57, 50, 40, 30, 19, 8],
    [15, 30, 45, 56, 64, 70, 73, 73, 71, 65, 57, 48, 39, 30, 18, 7],
    [14, 34, 49, 61, 69, 74, 76, 76, 74, 68, 61, 51, 40, 30, 19, 9],
    [35, 48, 59, 68, 75, 82, 84, 83, 80, 75, 67, 56, 45, 33, 21, 10],
    [27, 40, 53, 66, 79, 90, 94, 93, 88, 82, 72, 60, 47, 35, 22, 10],
    [10, 21, 34, 45, 60, 75, 90, 100, 102, 99, 88, 75, 59, 45, 30, 15],
    [4, 15, 26, 37, 48, 59, 70, 82, 93, 103, 106, 98, 85, 66, 45, 23],
    [5, 15, 25, 36, 46, 56, 67, 76, 85, 94, 102, 108, 103, 82, 58, 34],
    [6, 15, 24, 32, 40, 48, 55, 64, 72, 82, 94, 105, 115, 115, 83, 48],
    [4, 12, 20, 28, 36, 44, 53, 60, 68, 75, 83, 90, 98, 106, 111, 90],
    [9, 16, 23, 30, 37, 45, 51, 58, 65, 72, 80, 86, 94, 100, 106, 108],
    [9, 18, 27, 35, 44, 52, 59, 66, 74, 82, 87, 92, 95, 100, 96, 84],
    [4, 14, 23, 32, 41, 49, 57, 64, 71, 78, 84, 90, 94, 95, 83, 50],
    [5, 18, 30, 40, 48, 56, 64, 70, 77, 82, 85, 88, 89, 84, 64, 35],
    [7, 20, 32, 42, 52, 60, 69, 76, 82, 87, 89, 90, 83, 71, 54, 30],
    [7, 21, 34, 45, 57, 68, 75, 81, 84, 84, 83, 80, 72, 58, 44, 20],
    [8, 24, 36, 48, 58, 68, 76, 82, 85, 83, 78, 69, 59, 49, 34, 15],
    [13, 25, 36, 47, 57, 65, 70, 75, 76, 75, 71, 65, 57, 45, 30, 15],
    [10, 25, 38, 48, 57, 64, 69, 71, 72, 69, 64, 60, 51, 41, 29, 16],
    [7, 19, 30, 40, 48, 55, 59, 62, 62, 60, 55, 49, 41, 32, 23, 13],
    [5, 19, 29, 37, 42, 45, 46, 46, 45, 43, 39, 35, 30, 22, 14, 7],
    [0, 12, 17, 26, 34, 43, 49, 51, 54, 50, 46, 40, 32, 24, 14, 4],
    [2, 12, 20, 28, 35, 40, 45, 48, 51, 49, 45, 38, 32, 23, 15, 6],
    [10, 20, 29, 36, 42, 46, 49, 51, 52, 50, 45, 39, 32, 24, 15, 7],
    [8, 16, 26, 34, 41, 47, 49, 50, 50, 47, 42, 36, 29, 21, 12, 4],
    [9, 21, 32, 40, 49, 54, 55, 55, 52, 48, 43, 36, 29, 21, 13, 4],
    [12, 24, 34, 41, 46, 51, 55, 56, 51, 46, 40, 33, 27, 20, 13, 6],
    [14, 31, 42, 50, 55, 60, 60, 57, 50, 45, 39, 33, 26, 20, 13, 6],
    [10, 29, 45, 55, 60, 61, 60, 57, 53, 46, 40, 34, 25, 18, 11, 4],
    [20, 40, 60, 69, 71, 69, 65, 58, 50, 43, 36, 29, 24, 18, 12, 5],
    [30, 55, 72, 81, 79, 72, 64, 57, 50, 42, 35, 30, 24, 17, 12, 5],
    [62, 76, 85, 88, 85, 80, 71, 62, 55, 47, 41, 34, 27, 20, 14, 6],
    [60, 71, 79, 84, 85, 86, 82, 74, 66, 57, 48, 40, 31, 24, 16, 8],
    [20, 50, 72, 86, 89, 89, 86, 80, 72, 63, 54, 45, 36, 27, 18, 9],
    [26, 49, 63, 73, 82, 87, 87, 83, 78, 71, 62, 51, 40, 28, 18, 4],
    [15, 37, 52, 65, 73, 79, 82, 84, 79, 73, 63, 53, 40, 30, 17, 6]
]


def lch_sc_to_xyY(lch: Vector) -> Vector:
    """Convert from LCh to xy coordinates."""

    return util.xyz_to_xyY(lab_to_xyz(lch_to_lab(lch), XYZ_W), XYZ_W)


def to_lch_sc(color: Color) -> Vector:
    """Convert a color to LCh with an SC illuminant."""

    xyz = color.convert('xyz-d65').normalize(nans=False)
    xyz_sc = color.chromatic_adaptation(xyz._space.WHITE, WHITE_POINT_SC, xyz[:-1])
    return lab_to_lch(xyz_to_lab(xyz_sc, util.xy_to_xyz(WHITE_POINT_SC)))


def from_lch_sc(color: Color, lch: Vector) -> Color:
    """Convert a color from LCh with an SC illuminant."""

    xyz_sc = lab_to_xyz(lch_to_lab(lch), util.xy_to_xyz(WHITE_POINT_SC))
    xyz_d65 = color.chromatic_adaptation(WHITE_POINT_SC, color.CS_MAP['xyz-d65'].WHITE, xyz_sc)
    return color.update('xyz-d65', xyz_d65, color[-1])


def closest_lightness(l: float) -> tuple[int, float]:
    """Calculate the two closest lightness values and return the first index and interpolation factor."""

    # Handle too low lightness inside tolerance
    if l <= LCH_L[0]:
        li = 0
        lf = 0.0

    # Handle too high lightness inside tolerance
    elif l >= LCH_L[-1]:
        li = len(LCH_L) - 2
        lf = 1.0

    # Handle lightness with gamut
    else:
        li = bisect.bisect(LCH_L, l) - 1
        l1, l2 = LCH_L[li:li + 2]
        lf = 1 - (l2 - l) / (l2 - l1)

    return li, lf


def closest_hue(h: float) -> tuple[int, float]:
    """Calculate the two closest hues and return the first index and interpolation factor."""

    hi = 0
    # Handle wrapping hue
    if h > LCH_H[-1]:
        hi = len(LCH_H) - 1
        h1, h2 = 350, 360

    # Handle non-wrapping hue
    else:
        hi = bisect.bisect(LCH_H, h) - 1
        h1, h2 = LCH_H[hi:hi + 2]
    hf = 1 - (h2 - h) / (h2 - h1)

    return hi, hf


def get_chroma_limit(l: float, h: float) -> float:
    """Get the chroma limit."""

    # Find the two closest lightness columns and calculate the needed interpolation factor.
    li, lf = closest_lightness(l)

    # Find the two closest hue rows and calculate the needed interpolation factor.
    hi, hf = closest_hue(h)

    # Interpolate the chroma limit by interpolating chroma values for the closest lightness values and hues.
    if hi == len(LCH_H) - 1:
        row1, row2 = LCH_POINTER[-1], LCH_POINTER[0]
    else:
        row1, row2 = LCH_POINTER[hi:hi + 2]
    return alg.lerp(alg.lerp(row1[li], row1[li + 1], lf), alg.lerp(row2[li], row2[li + 1], lf), hf)


def fit_pointer_gamut(color: Color) -> Color:
    """Fit a color to the Pointer gamut."""

    # Convert to CIE LCh with the SC illuminant
    l, c, h = to_lch_sc(color)

    # Clamp lightness
    new_l = max(LCH_L[0], l)
    new_l = min(LCH_L[-1], new_l)

    new_c = min(c, get_chroma_limit(l, h))

    adjusted = l != new_l or c != new_c

    # Adjust original color only if a modification was made
    return from_lch_sc(color, [new_l, new_c, h]) if adjusted else color


def in_pointer_gamut(color: Color, tolerance: float) -> bool:
    """
    See if color is within the pointer gamut.

    Convert to CIE LCh using the SC illuminant.
    Find the closest hues and lightness (rows and columns) so we can interpolate
    the an appropriate max chroma for a given hue and lightness. Test that the
    color's chroma does not exceed the limit.
    """

    # Convert to CIE LCh with the SC illuminant
    l, c, h = to_lch_sc(color)

    # If lightness exceeds the acceptable range, then we are not in gamut
    if (l < (LCH_L[0] - tolerance)) or (l > (LCH_L[-1] + tolerance)):
        return False

    # Test that the color does not exceed the max chroma
    return c <= (get_chroma_limit(l, h) + tolerance)


def pointer_gamut_boundary(lightness: float | None = None) -> list[Vector]:
    """
    Calculate the Pointer gamut boundary points for the given lightness.

    If no lightness is provided, calculate the maximum pointer boundary.
    Result is returned as xyY coordinates (C illuminant).
    """

    # Maximum Pointer gamut boundary
    # For each hue, find the lightness/chroma point that is furthest away from the white point.
    if lightness is None:
        max_gamut = []  # type: list[Vector]
        for i, h in enumerate(LCH_H):
            max_dxy = 0.0
            max_xyy = [0.0, 0.0, 0.0]
            for j, l in enumerate(LCH_L):
                xyy = lch_sc_to_xyY([l, LCH_POINTER[i][j], h])
                dxy = math.sqrt((WHITE_POINT_SC[0] - xyy[0]) ** 2 + (WHITE_POINT_SC[1] - xyy[1]) ** 2)
                if dxy > max_dxy:
                    max_dxy = dxy
                    max_xyy = xyy
            max_gamut.append(max_xyy)
        return max_gamut

    # Pointer gamut boundary at a given lightness
    # Return all the points for a given lightness
    elif LCH_L[0] <= lightness <= LCH_L[-1]:
        li, lf = closest_lightness(lightness)
        chroma = [alg.lerp(row[li], row[li + 1], lf) for row in LCH_POINTER]
        return [lch_sc_to_xyY([lightness, c, h]) for c, h in zip(chroma, LCH_H)]

    # Lightness exceeds threshold
    else:
        raise ValueError('Lightness must be between {} and {}, but was {}'.format(LCH_L[0], LCH_L[-1], lightness))



"""Test sRGB Linear library."""
import unittest
from . import util
from coloraide import Color
import pytest


class TestsRGBLinear(util.ColorAssertsPyTest):
    """Test sRGB Linear."""

    COLORS = [
        ('red', 'color(srgb-linear 1 0 0)'),
        ('orange', 'color(srgb-linear 1 0.37626 0)'),
        ('yellow', 'color(srgb-linear 1 1 0)'),
        ('green', 'color(srgb-linear 0 0.21586 0)'),
        ('blue', 'color(srgb-linear 0 0 1)'),
        ('indigo', 'color(srgb-linear 0.07036 0 0.22323)'),
        ('violet', 'color(srgb-linear 0.85499 0.22323 0.85499)'),
        ('white', 'color(srgb-linear 1 1 1)'),
        ('gray', 'color(srgb-linear 0.21586 0.21586 0.21586)'),
        ('black', 'color(srgb-linear 0 0 0)'),
        # Test CSS color
        ('color(srgb-linear 0 0.50196 0)', 'color(srgb-linear 0 0.50196 0)'),
        ('color(srgb-linear 0 0.50196 0 / 0.5)', 'color(srgb-linear 0 0.50196 0 / 0.5)'),
        ('color(srgb-linear 50% 50% 50% / 50%)', 'color(srgb-linear 0.5 0.5 0.5 / 0.5)'),
        ('color(srgb-linear none none none / none)', 'color(srgb-linear none none none / none)'),
        # Test range
        ('color(srgb-linear 0% 0% 0%)', 'color(srgb-linear 0 0 0)'),
        ('color(srgb-linear 100% 100% 100%)', 'color(srgb-linear 1 1 1)'),
        ('color(srgb-linear -100% -100% -100%)', 'color(srgb-linear -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('srgb-linear'), Color(color2))


class TestLinearsRGBSerialize(util.ColorAssertsPyTest):
    """Test Linear sRGB serialization."""

    COLORS = [
        # Test color
        ('color(srgb-linear 0 0.3 0.75 / 0.5)', {}, 'color(srgb-linear 0 0.3 0.75 / 0.5)'),
        # Test alpha
        ('color(srgb-linear 0 0.3 0.75)', {'alpha': True}, 'color(srgb-linear 0 0.3 0.75 / 1)'),
        ('color(srgb-linear 0 0.3 0.75 / 0.5)', {'alpha': False}, 'color(srgb-linear 0 0.3 0.75)'),
        # Test None
        ('color(srgb-linear none 0.3 0.75)', {}, 'color(srgb-linear 0 0.3 0.75)'),
        ('color(srgb-linear none 0.3 0.75)', {'none': True}, 'color(srgb-linear none 0.3 0.75)'),
        # Test Fit
        ('color(srgb-linear 1.2 0.2 0)', {}, 'color(srgb-linear 1 0.22803 0.02349)'),
        ('color(srgb-linear 1.2 0.2 0)', {'fit': False}, 'color(srgb-linear 1.2 0.2 0)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestsRGBLinearProperties(util.ColorAsserts, unittest.TestCase):
    """Test sRGB Linear properties."""

    def test_red(self):
        """Test `red`."""

        c = Color('color(srgb-linear 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['red'], 0.1)
        c['red'] = 0.2
        self.assertEqual(c['red'], 0.2)

    def test_green(self):
        """Test `green`."""

        c = Color('color(srgb-linear 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['green'], 0.2)
        c['green'] = 0.1
        self.assertEqual(c['green'], 0.1)

    def test_blue(self):
        """Test `blue`."""

        c = Color('color(srgb-linear 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['blue'], 0.3)
        c['blue'] = 0.1
        self.assertEqual(c['blue'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(srgb-linear 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Test filter method."""
from coloraide import Color
from . import util
import pytest
import unittest


class TestFilters(util.ColorAssertsPyTest):
    """Test filters."""

    @pytest.mark.parametrize(
        'color1,color2,name,amount',
        [
            ('red', 'rgb(168.28 159.48 142.36)', 'sepia', 1),
            ('red', 'rgb(217.36 115.95 103.13)', 'sepia', 0.5),
            ('red', 'rgb(255 0 0)', 'sepia', 0),
            ('red', 'rgb(255 0 0)', 'sepia', -1),
            ('red', 'rgb(168.28 159.48 142.36)', 'sepia', 2),

            ('orange', 'rgb(184.41 184.41 184.41)', 'grayscale', 1),
            ('orange', 'rgb(223.39 175.05 134.63)', 'grayscale', 0.5),
            ('orange', 'rgb(255 165 0)', 'grayscale', 0),
            ('orange', 'rgb(255 165 0)', 'grayscale', -1),
            ('orange', 'rgb(184.41 184.41 184.41)', 'grayscale', 2),

            ('yellow', 'rgb(255 255 0)', 'saturate', 1),
            ('yellow', 'rgb(250.92 250.92 181.34)', 'saturate', 0.5),
            ('yellow', 'rgb(246.75 246.75 246.75)', 'saturate', 0),
            ('yellow', 'rgb(246.75 246.75 246.75)', 'saturate', -1),
            ('yellow', 'rgb(262.91 262.91 -246.75)', 'saturate', 2),

            ('green', 'rgb(255 229.08 255)', 'invert', 1),
            ('green', 'rgb(187.52 187.52 187.52)', 'invert', 0.5),
            ('green', 'rgb(0 128 0)', 'invert', 0),
            ('green', 'rgb(0 128 0)', 'invert', -1),
            ('green', 'rgb(255 229.08 255)', 'invert', 2),

            ('blue', 'rgb(0 0 255)', 'brightness', 1),
            ('blue', 'rgb(0 0 187.52)', 'brightness', 0.5),
            ('blue', 'rgb(0 0 0)', 'brightness', 0),
            ('blue', 'rgb(0 0 0)', 'brightness', -1),
            ('blue', 'rgb(0 0 345.08)', 'brightness', 2),

            ('indigo', 'rgb(75 0 130)', 'contrast', 1),
            ('indigo', 'rgb(145.47 136.96 162.06)', 'contrast', 0.5),
            ('indigo', 'rgb(187.52 187.52 187.52)', 'contrast', 0),
            ('indigo', 'rgb(187.52 187.52 187.52)', 'contrast', -1),
            ('indigo', 'rgb(-161.59 -187.52 -65.424)', 'contrast', 2),

            ('violet', 'rgb(238 130 238)', 'hue-rotate', 0),
            ('violet', 'rgb(238 152.19 -62.17)', 'hue-rotate', 90),
            ('violet', 'rgb(-62.17 200.89 -62.17)', 'hue-rotate', 180),
            ('violet', 'rgb(-62.17 186.12 238)', 'hue-rotate', 270),
            ('violet', 'rgb(238 130 238)', 'hue-rotate', 360),
            ('violet', 'rgb(238 152.19 -62.17)', 'hue-rotate', 450),
            ('violet', 'rgb(-62.17 186.12 238)', 'hue-rotate', -90),

            ('black', 'rgb(0 0 0)', 'opacity', 1),
            ('black', 'rgb(0 0 0 / 0.5)', 'opacity', 0.5),
            ('black', 'rgb(0 0 0 / 0)', 'opacity', 0),
            ('black', 'rgb(0 0 0 / 0)', 'opacity', -1),
            ('black', 'rgb(0 0 0)', 'opacity', 2)
        ]
    )
    def test_filter(self, color1, color2, name, amount):
        """Test filter."""

        print('color1: ', color1)
        print('color2: ', color2)
        self.assertColorEqual(
            Color(color1).filter(name, amount, out_space='srgb'),
            Color(color2) if isinstance(color2, str) else color2
        )


class TestFilterMisc(util.ColorAsserts, unittest.TestCase):
    """Test filter miscellaneous cases."""

    def test_srgb(self):
        """Test using a filter with sRGB."""

        self.assertColorEqual(Color('red').filter('sepia', space='srgb'), Color('rgb(100.22 88.995 69.36)'))
        self.assertColorNotEqual(Color('red').filter('sepia'), Color('red').filter('sepia', space='srgb'))

    def test_bad_space(self):
        """Test a bad input space."""

        with self.assertRaises(ValueError):
            Color('red').filter('sepia', space='display-p3', out_space='srgb')

    def test_bad_filter(self):
        """Test bad filter."""

        with self.assertRaises(ValueError):
            Color('red').filter('bad')



"""
JzCzhz class.

https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
"""
from __future__ import annotations
import math
from ..cat import WHITES
from .lch import LCh
from .jzazbz import Jzazbz
from ..channels import Channel, FLG_ANGLE
from ..types import Vector


class JzCzhz(LCh):
    """
    JzCzhz class.

    https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
    """

    BASE = "jzazbz"
    NAME = "jzczhz"
    SERIALIZE = ("--jzczhz",)
    WHITE = WHITES['2deg']['D65']
    DYNAMIC_RANGE = 'hdr'
    CHANNEL_ALIASES = {
        "lightness": "jz",
        "chroma": "cz",
        "hue": "hz"
    }
    ACHROMATIC = Jzazbz.ACHROMATIC
    CHANNELS = (
        Channel("jz", 0.0, 1.0, limit=(0.0, None)),
        Channel("cz", 0.0, 0.5, limit=(0.0, None)),
        Channel("hz", 0.0, 360.0, flags=FLG_ANGLE, nans=ACHROMATIC.hue)
    )

    def resolve_channel(self, index: int, coords: Vector) -> float:
        """Resolve channels."""

        if index == 2:
            h = coords[2]
            return self.ACHROMATIC.get_ideal_hue(coords[0]) if math.isnan(h) else h

        elif index == 1:
            c = coords[1]
            return self.ACHROMATIC.get_ideal_chroma(coords[0]) if math.isnan(c) else c

        value = coords[index]
        return self.channels[index].nans if math.isnan(value) else value

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return coords[0] == 0.0 or self.ACHROMATIC.test(*coords)

    def hue_name(self) -> str:
        """Hue name."""

        return "hz"



"""
B-Spline interpolation.

https://en.wikipedia.org/wiki/B-spline
https://www.math.ucla.edu/~baker/149.1.02w/handouts/dd_splines.pdf
http://www2.cs.uregina.ca/~anima/408/Notes/Interpolation/UniformBSpline.htm
"""
from __future__ import annotations
from .. import algebra as alg
from .continuous import InterpolatorContinuous
from ..interpolate import Interpolator, Interpolate
from ..types import Vector
from typing import Callable, Mapping, Sequence, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class InterpolatorBSpline(InterpolatorContinuous):
    """Interpolate with B-spline."""

    def adjust_endpoints(self) -> None:
        """Adjust endpoints such that they are clamped and can handle extrapolation."""

        # We cannot interpolate all the way to `coord[0]` and `coord[-1]` without additional control
        # points to coax the curve through the end points. Generate a point at both ends so that we
        # can properly evaluate the spline from start to finish. Additionally, when the extrapolating
        # past the 0 - 1 boundary, provide some linear behavior
        self.extrapolated = [
            list(zip(self.coordinates[0], self.coordinates[1])),
            list(zip(self.coordinates[-2], self.coordinates[-1]))
        ]
        self.coordinates.insert(0, [2 * a - b for a, b in zip(self.coordinates[0], self.coordinates[1])])
        self.coordinates.append([2 * a - b for a, b in zip(self.coordinates[-1], self.coordinates[-2])])

    def setup(self) -> None:
        """Optional setup."""

        # Process undefined values
        self.spline = alg.bspline
        self.handle_undefined()
        self.adjust_endpoints()

    def interpolate(
        self,
        point: float,
        index: int
    ) -> Vector:
        """Interpolate."""

        # Prepare in-boundary coordinates
        coords = list(zip(*self.coordinates[index - 1:index + 3]))

        # Apply interpolation to each channel
        channels = []
        for i in range(len(self.coordinates[0])):

            t = self.ease(point, i)

            # If `t` ends up spilling out past our boundaries, we need to extrapolate
            if self.extrapolate and index == 1 and point < 0.0:
                p0, p1 = self.extrapolated[0][i]
                channels.append(alg.lerp(p0, p1, t))
            elif self.extrapolate and index == self.length - 1 and point > 1.0:
                p0, p1 = self.extrapolated[1][i]
                channels.append(alg.lerp(p0, p1, t))
            else:
                p0, p1, p2, p3 = coords[i]
                channels.append(self.spline(p0, p1, p2, p3, t))

        # Small adjustment for floating point math and alpha channels
        if 1 - channels[-1] < 1e-6:
            channels[-1] = 1

        return channels


class BSpline(Interpolate):
    """B-spline interpolation plugin."""

    NAME = "bspline"

    def interpolator(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Return the B-spline interpolator."""

        return InterpolatorBSpline(
            coordinates,
            channel_names,
            create,
            easings,
            stops,
            space,
            out_space,
            progress,
            premultiplied,
            extrapolate,
            domain,
            padding,
            **kwargs
        )



"""
RLAB.

https://scholarworks.rit.edu/cgi/viewcontent.cgi?article=1153&context=article
https://www.imaging.org/site/PDFS/Papers/1997/RP-0-67/2368.pdf
"""
from __future__ import annotations
from ..cat import WHITES
from ..spaces.lab import Lab
from .. import algebra as alg
from ..types import Vector
from ..channels import Channel, FLG_MIRROR_PERCENT

XYZ_TO_XYZ_REF = [
    [1.0521266389510715, 2.220446049250313e-16, 0.0],
    [0.0, 1.0, 2.414043899674756e-19],
    [0.0, 0.0, 0.9182249511582473]
]

XYZ_REF_TO_XYZ = [
    [0.9504559270516716, -2.110436108208428e-16, 5.548406636355788e-35],
    [0.0, 1.0, -2.629033219615395e-19],
    [0.0, 0.0, 1.0890577507598784]
]

EXP = 2.3


def rlab_to_xyz(rlab: Vector) -> Vector:
    """RLAB to XYZ."""

    l, a, b = rlab
    yr = l / 100
    xr = alg.npow((a / 430) + yr, EXP)
    zr = alg.npow(yr - (b / 170), EXP)
    return alg.dot(XYZ_REF_TO_XYZ, [xr, alg.npow(yr, EXP), zr], dims=alg.D2_D1)


def xyz_to_rlab(xyz: Vector) -> Vector:
    """XYZ to RLAB."""

    xyz_ref = alg.dot(XYZ_TO_XYZ_REF, xyz, dims=alg.D2_D1)
    xr, yr, zr = [alg.nth_root(c, EXP) for c in xyz_ref]
    l = 100 * yr
    a = 430 * (xr - yr)
    b = 170 * (yr - zr)
    return [l, a, b]


class RLAB(Lab):
    """RLAB class."""

    BASE = 'xyz-d65'
    NAME = "rlab"
    SERIALIZE = ("--rlab",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("a", -125.0, 125.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -125.0, 125.0, flags=FLG_MIRROR_PERCENT)
    )

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Hunter Lab."""

        return rlab_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Hunter Lab."""

        return xyz_to_rlab(coords)



"""Lab class."""
from __future__ import annotations
from .. import lab as base
from ...css import parse
from ...css import serialize
from ...types import Vector
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ...color import Color


class Lab(base.Lab):
    """Lab class."""

    def to_string(
        self,
        parent: Color,
        *,
        alpha: bool | None = None,
        precision: int | None = None,
        fit: str | bool = True,
        none: bool = False,
        color: bool = False,
        percent: bool = False,
        **kwargs: Any
    ) -> str:
        """Convert to CSS."""

        return serialize.serialize_css(
            parent,
            func='lab',
            alpha=alpha,
            precision=precision,
            fit=fit,
            none=none,
            color=color,
            percent=percent
        )

    def match(
        self,
        string: str,
        start: int = 0,
        fullmatch: bool = True
    ) -> tuple[tuple[Vector, float], int] | None:
        """Match a CSS color string."""

        return parse.parse_css(self, string, start, fullmatch)



"""
ACEScg color space.

https://www.oscars.org/science-technology/aces/aces-documentation
"""
from __future__ import annotations
from ..channels import Channel
from ..spaces.srgb import sRGB
from .. import algebra as alg
from ..types import Vector

AP1_TO_XYZ = [
    [0.6624541811085053, 0.13400420645643313, 0.1561876870049078],
    [0.27222871678091454, 0.6740817658111484, 0.05368951740793705],
    [-0.005574649490394108, 0.004060733528982826, 1.0103391003129971]
]

XYZ_TO_AP1 = [
    [1.6410233796943257, -0.32480329418479, -0.23642469523761225],
    [-0.6636628587229829, 1.6153315916573379, 0.016756347685530137],
    [0.011721894328375376, -0.008284441996237409, 0.9883948585390215]
]


def acescg_to_xyz(acescg: Vector) -> Vector:
    """Convert ACEScc to XYZ."""

    return alg.dot(AP1_TO_XYZ, acescg, dims=alg.D2_D1)


def xyz_to_acescg(xyz: Vector) -> Vector:
    """Convert XYZ to ACEScc."""

    return alg.dot(XYZ_TO_AP1, xyz, dims=alg.D2_D1)


class ACEScg(sRGB):
    """The ACEScg color class."""

    BASE = "xyz-d65"
    NAME = "acescg"
    SERIALIZE = ("--acescg",)  # type: tuple[str, ...]
    WHITE = (0.32168, 0.33767)
    CHANNELS = (
        Channel("r", 0.0, 65504.0, bound=True),
        Channel("g", 0.0, 65504.0, bound=True),
        Channel("b", 0.0, 65504.0, bound=True)
    )
    DYNAMIC_RANGE = 'hdr'

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return acescg_to_xyz(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return xyz_to_acescg(coords)



"""
Uncalibrated, naive CMYK color space.

https://www.w3.org/TR/css-color-5/#cmyk-rgb
"""
from __future__ import annotations
from ..spaces import Space
from ..channels import Channel
from ..cat import WHITES
from ..types import Vector
from .. import algebra as alg
import math


def srgb_to_cmyk(rgb: Vector) -> Vector:
    """Convert sRGB to CMYK."""

    k = 1.0 - max(rgb)
    c = m = y = 0.0
    if k != 1:
        r, g, b = rgb
        c = (1.0 - r - k) / (1.0 - k)
        m = (1.0 - g - k) / (1.0 - k)
        y = (1.0 - b - k) / (1.0 - k)

    return [c, m, y, k]


def cmyk_to_srgb(cmyk: Vector) -> Vector:
    """Convert CMYK to sRGB."""

    c, m, y, k = cmyk
    return [
        1.0 - min(1.0, c * (1.0 - k) + k),
        1.0 - min(1.0, m * (1.0 - k) + k),
        1.0 - min(1.0, y * (1.0 - k) + k)
    ]


class CMYK(Space):
    """The CMYK color class."""

    BASE = "srgb"
    NAME = "cmyk"
    SERIALIZE = ("--cmyk",)  # type: tuple[str, ...]
    CHANNELS = (
        Channel("c", 0.0, 1.0, bound=True),
        Channel("m", 0.0, 1.0, bound=True),
        Channel("y", 0.0, 1.0, bound=True),
        Channel("k", 0.0, 1.0, bound=True)
    )
    CHANNEL_ALIASES = {
        "cyan": 'c',
        "magenta": 'm',
        "yellow": 'y',
        "black": 'k'
    }
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Test if color is achromatic."""

        if math.isclose(1.0, coords[-1], abs_tol=1e-4):
            return True

        black = [1, 1, 1]
        for x in alg.vcross(coords[:-1], black):
            if not math.isclose(0.0, x, abs_tol=1e-5):
                return False
        return True

    def to_base(self, coords: Vector) -> Vector:
        """To sRGB."""

        return cmyk_to_srgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From sRGB."""

        return srgb_to_cmyk(coords)
